# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""XML renderer"""

from __future__ import with_statement

import types
import copy
import cStringIO
import urllib

import peak.rules

from lxml import etree as ET

from nagare.namespaces import common

CHECK_ATTRIBUTES = False

# Namespace for the ``meld:id`` attribute
_MELD_NS = 'http://www.plope.com/software/meld3'
_MELD_ID = '{%s}id' % _MELD_NS


# ---------------------------------------------------------------------------

class _Tag(ET.ElementBase):
    """A xml tag
    """
    def init(self, renderer):
        """Each tag keeps track of the renderer that created it

        Return:
           - ``self``
        """
        self._renderer = renderer
        return self

    @property
    def renderer(self):
        """Return the renderer that created this tag

        Return:
          - the renderer
        """
        # The renderer is search first, in this tag, else at the root of the tree
        return getattr(self, '_renderer', None) or self.getroottree().getroot()._renderer

    def write_xmlstring(self, encoding='utf-8', pipeline=True, **kw):
        """Serialize in XML the tree beginning at this tag

        In:
          - ``encoding`` -- encoding of the XML
          - ``pipeline`` -- if False, the ``meld:id`` attributes are deleted

        Return:
          - the XML
        """
        if not pipeline:
            for element in self.xpath('.//*[@meld:id]', namespaces={'meld': _MELD_NS}):
                del element.attrib[_MELD_ID]

        return ET.tostring(self, encoding=encoding, method='xml', **kw)

    def xpath(self, *args, **kw):
        """Override ``xpath()`` to associate a renderer to all the returned nodes
        """
        nodes = super(_Tag, self).xpath(*args, **kw)

        renderer = self.renderer
        for node in nodes:
            node._renderer = renderer

        return nodes

    def findmeld(self, id, default=None):
        """Find a tag with a given ``meld:id`` value

        In:
          - ``id`` -- value of the ``meld:id`` attribute to search
          - ``default`` -- value returned if the tag is not found

        Return:
          - the tag found, else the ``default`` value
        """
        nodes = self.xpath('.//*[@meld:id="%s"]' % id, namespaces={'meld': _MELD_NS})

        if len(nodes) != 0:
            # Return only the first tag found
            return nodes[0]

        return default

    def append_text(self, text):
        """Append a text to this tag

        In:
          - ``text`` -- text to add
        """
        if len(self) != 0:
            self[-1].tail = (self[-1].tail or '') + text
        else:
            self.text = (self.text or '') + text

    def add_child(self, child):
        """Append a child to this tag

        In:
          - ``child`` -- child to add
        """
        # Forward the call to the generic method
        add_child(self, child)

    def meld_id(self, id):
        """Set the value of the attribute ``meld:id`` of this tag

        In:
          - ``id`` - value of the ``meld;id`` attribute

        Return:
          - ``self``
        """
        self.add_child({_MELD_ID: id})
        return self

    def fill(self, *children, **attrib):
        """Change all the child and append attributes of this tag

        In:
          - ``children`` -- list of the new children of this tag
          - ``attrib`` -- dictionnary of attributes of this tag

        Return:
          - ``self``
        """
        self[:] = []
        self.text = None

        return self.__call__(*children, **attrib)

    def replace(self, *children):
        """Replace this tag by others

        In:
          - ``children`` -- list of tags to replace this tag
        """
        parent = self.getparent()

        # We can not replace the root of the tree
        if parent is not None:
            tail = self.tail
            l = len(parent)

            i = parent.index(self)
            parent.fill(parent.text or '', parent[:i], children, self.tail or '', parent[i + 1:])

            if len(parent) >= l:
                parent[i - l].tail = tail

    def repeat(self, iterable, childname=None):
        """Iterate over a sequence, cloning a new child each time

        In:
          - ``iterable`` -- the sequence
          - ``childname`` -- If ``None``, clone this tag each time else
            clone each time the tag that have ``childname`` as ``meld:id`` value

        Return:
          - list of tuples (cloned child, item of the sequence)
        """
        # Find the child to clone
        if childname is None:
            element = self
        else:
            element = self.findmeld(childname)

        parent = element.getparent()
        parent.remove(element)

        for thing in iterable:
            clone = copy.deepcopy(element)
            clone._renderer = element.renderer
            parent.append(clone)

            yield (clone, thing)

    def __call__(self, *children, **attrib):
        """Append child and attributes to this tag

        In:
          - ``children`` -- children to add
          - ``attrib`` -- attributes to add

        Return:
          - ``self``
        """
        if attrib:
            self.add_child(attrib)

        nb = len(children)
        if nb != 0:
            if nb == 1:
                children = children[0]
            self.add_child(children)

        if CHECK_ATTRIBUTES and self._authorized_attribs and not frozenset(self.attrib).issubset(self._authorized_attribs):
            raise AttributeError("Bad attributes for element <%s>: " % self.tag + ', '.join(list(frozenset(self.attrib) - self._authorized_attribs)))

        return self

    def __enter__(self):
        return self.renderer.enter(self)

    def __exit__(self, exception, data, tb):
        if exception is None:
            self.renderer.exit()


# ---------------------------------------------------------------------------

# Generic methods to add a child to a tag
# ---------------------------------------

def add_child(self, o):
    """Default method to add an object to a tag

    In:
      - ``self`` -- the tag
      - ``o`` -- object to add

    Try to add the result of ``o.render()`` to the tag
    """
    render = getattr(o, 'render', None)
    if render is None:
        raise TypeError("Can't append a '%s' to element <%s>" % (type(o), self.tag))

    self.add_child(render(self.renderer))


@peak.rules.when(add_child, (_Tag, basestring))
def add_child(self, s):
    """Add a string to a tag

    In:
      - ``self`` -- the tag
      - ``s`` - str or unicode string to add
    """
    self.append_text(s)


@peak.rules.when(add_child, (_Tag, (list, tuple, types.GeneratorType)))
def add_child(self, i):
    """Add elements to a tag

    In:
      - ``self`` -- the tag
      - ``i`` -- elements to add

    Add each element as a child
    """
    for child in i:
        self.add_child(child)


@peak.rules.when(add_child, (_Tag, (int, long, float)))
def add_child(self, n):
    """Add a number to a tag

    In:
      - ``self`` -- the tag
      - ``n`` -- number to add

    Convert the number to string and then add it
    """
    self.append_text(str(n))


@peak.rules.when(add_child, (_Tag, ET._Element))
def add_child(self, element):
    """Add a tag to a tag

    In:
      - ``self`` -- the tag
      - ``element`` -- the tag to add
    """
    if hasattr(element, '_renderer'):
        del element._renderer

    self.append(element)


@peak.rules.when(add_child, (_Tag, ET._Comment))
def add_child(self, element):
    """Add a comment element to a tag

    In:
      - ``self`` -- the tag
      - ``element`` -- the comment to add
    """
    if hasattr(element, '_renderer'):
        del element._renderer

    self.append(element)


@peak.rules.when(add_child, (_Tag, ET._ProcessingInstruction))
def add_child(self, element):
    """Add a PI element to a tag

    In:
      - ``self`` -- the tag
      - ``element`` -- the PI to add

    Do nothing
    """
    self.append(element)


@peak.rules.when(add_child, (_Tag, dict))
def add_child(self, d):
    """Add a dictionary to a tag

    In:
      - ``self`` -- the tag
      - ``element`` -- the dictionary to add

    Each key/value becomes an attribute of the tag

    Attribute name can end with a '_' which is removed
    """
    for (name, value) in d.items():
        add_attribute(self, name, value)


# ---------------------------------------------------------------------------

# Generic methods to add an attribute to a tag
# --------------------------------------------

def add_attribute(self, name, value):
    """Default method to add an attribute to a tag

    In:
      - ``self`` -- the tag
      - ``name`` -- name of the attribute to add
      - ``value`` -- value of the attribute to add
    """
    add_attribute(self, name, unicode(value))


@peak.rules.when(add_attribute, (_Tag, basestring, basestring))
def add_attribute(self, name, value):
    if name.endswith('_'):
        name = name[:-1]

    self.set(name, value)


@peak.rules.when(add_attribute, (ET._ProcessingInstruction, basestring, basestring))
def add_attribute(self, name, value):
    if name.endswith('_'):
        name = name[:-1]

    self.text += (' %s="%s"' % (name, value))


# ---------------------------------------------------------------------------

class TagProp(object):
    """Tag factory with a behavior of an object attribute

    Each time this attribute is read, a new tag is created
    """
    def __init__(self, name, authorized_attribs=None, factory=None):
        """Initialization

        In:
          - ``name`` -- name of the tags to create
          - ``authorized_attribs`` -- names of the valid attributes
          - ``factory`` -- special factory to create the tag
        """
        self._name = name
        self._factory = factory

        if CHECK_ATTRIBUTES:
            self._authorized_attribs = frozenset(authorized_attribs) if authorized_attribs is not None else None

    def __get__(self, renderer, cls):
        """Create a new tag each time this attribute is read

        In:
          - ``renderer`` -- the object that has this attribute
          - ``cls`` -- *not used*

        Return:
          - a new tag
        """
        element = renderer.makeelement(self._name)

        if CHECK_ATTRIBUTES:
            element._authorized_attribs = self._authorized_attribs

        return element


class RendererMetaClass(type):
    """Meta class for the renderer class

    Discover the tags that have a special factory and pass them to the
    ``class_init()`` method of the class
    """
    def __new__(self, cls_name, bases, ns):
        """Creation of the class

        In:
          - ``cls_name`` -- name of the class to create
          - ``bases`` -- tuple of the base classes of the class to create
          - ``ns`` -- namespace of the class to create

        Return:
          - the new class
        """
        cls = super(RendererMetaClass, self).__new__(self, cls_name, bases, ns)

        special_tags = {}

        for (k, v) in ns.items():
            if isinstance(v, TagProp):
                if v._factory is not None:
                    special_tags[v._name] = v._factory
                del v._factory

        cls.class_init(special_tags)

        return cls


# -----------------------------------------------------------------------

class XmlRenderer(common.Renderer):
    """The base class of all the renderers that generate a XML dialect
    """
    __metaclass__ = RendererMetaClass

    doctype = ''
    content_type = 'text/xml'

    @classmethod
    def class_init(cls, special_tags):
        """Class initialisation

        In:
          -- ``special_tags`` -- tags that have a special factory
        """
        # Create a XML parser that generate ``_Tag`` nodes
        cls._xml_parser = ET.XMLParser()
        cls._xml_parser.setElementClassLookup(ET.ElementDefaultClassLookup(element=_Tag))

    def __init__(self, parent=None):
        """Renderer initialisation

        In:
          - ``parent`` -- parent renderer
        """
        if parent is None:
            self.namespaces = None
            self._default_namespace = None
        else:
            # This renderer use the same XML namespaces than its parent
            self.namespaces = parent.namespaces
            self._default_namespace = parent._default_namespace

        self.parent = parent
        self._prefix = ''

        # The stack, contening the last tag push by a ``with`` statement
        # Initialized with a dummy root
        self._stack = [self.makeelement('_renderer_root_')]

        # Each renderer created has a unique id
        self.id = self.generate_id('id')

    def _get_default_namespace(self):
        """Return the default_namespace

        Return:
          - the default namespace or ``None`` if no default namespace was set
        """
        return self._default_namespace

    def _set_default_namespace(self, namespace):
        """Change the default namespace

        The namespace must be a key of the ``self.namespaces`` dictionary or be
        ``None``

        For example:

        .. code-block:: python

          x.namespaces = { 'xsl' : 'http://www.w3.org/1999/XSL/Transform' }
          x.set_default_namespace('xsl')
        """
        self._default_namespace = namespace

        if namespace is None:
            self._prefix = ''
        else:
            self._prefix = '{%s}' % self.namespaces[namespace]
    default_namespace = property(_get_default_namespace, _set_default_namespace)

    @property
    def root(self):
        """Return the first tag(s) sent to the renderer

        .. warning::
            A list of tags can be returned

        Return:
          - the tag(s)
        """
        children = self._stack[0].getchildren()

        text = self._stack[0].text

        if text is not None:
            children.insert(0, text)

        if not children:
            return ''

        if len(children) == 1:
            return children[0]

        return children

    def enter(self, current):
        """A new tag is pushed by a ``with`` statement

        In:
          - ``current`` -- the tag

        Return:
          - ``current``
        """
        self << current  # Append the tag to its parent
        self._stack.append(current)  # Push it onto the stack

        return current

    def exit(self):
        """End of a ``with`` statement
        """
        self._stack.pop()   # Pop the current tag from the stack

    def _makeelement(self, tag, parser):
        """Make a tag, in the default namespace set

        In:
          - ``tag`` -- name of the tag to create
          - ``parser`` -- parser used to create the tag

        Return:
          - the new tag
        """
        # 1. Create the tag with in the default namespace
        element = parser.makeelement(self._prefix + tag, nsmap=self.namespaces)
        # 2. Initialize it with this renderer
        return element.init(self)

    def makeelement(self, tag):
        """Make a tag, in the default namespace set

        In:
          - ``tag`` -- name of the tag to create

        Return:
          - the new tag
        """
        # Create the tag with the XML parser
        return self._makeelement(tag, self._xml_parser)

    def __lshift__(self, current):
        """Add a tag to the last tag pushed by a ``with`` statement

        In:
          - ``current`` -- tag to add

        Return:
          - ``self``, the renderer
        """
        add_child(self._stack[-1], current)
        return self

    def comment(self, text=''):
        """Create a comment element

        In:
          - ``text`` -- text of the comment

        Return:
          - the new comment element
        """
        return ET.Comment(text)

    def processing_instruction(self, target, text=None, **kw):
        """Create a processing instruction element

        In:
          - ``target`` -- the PI target
          - ``text`` -- the PI text
          - ``kw`` -- pseudo attributes

        Return:
          - the new processing instruction element
        """
        pi = ET.ProcessingInstruction(target, text)

        for name, value in kw.items():
            add_attribute(pi, name, value)

        return pi

    def parse_xml(self, source, fragment=False, no_leading_text=False, **kw):
        """Parse a XML file

        In:
          - ``source`` -- can be a filename or a file object
          - ``fragment`` -- if ``True``, can parse a XML fragment i.e a XML without
            a unique root
          - ``no_leading_text`` -- if ``fragment`` is ``True``, ``no_leading_text``
            is ``False`` and the XML to parsed begins by a text, this text is keeped
          - ``kw`` -- keywords parameters are passed to the XML parser

        Return:
          - the root element of the parsed XML, if ``fragment`` is ``False``
          - a list of XML elements, if ``fragment`` is ``True``
        """
        if isinstance(source, basestring):
            if source.startswith(('http://', 'https://', 'ftp://')):
                source = urllib.urlopen(source)
            else:
                source = open(source)

        # Create a dedicated XML parser with the ``kw`` parameter
        parser = ET.XMLParser(**kw)
        # This parser will generate nodes of type ``_Tag``
        parser.setElementClassLookup(ET.ElementDefaultClassLookup(element=_Tag))

        if not fragment:
            # Parse a XML file
            # ----------------

            root = ET.parse(source, parser).getroot()
            source.close()

            # Attach the renderer to the root
            root._renderer = self
            return root

        # Parse a XML fragment
        # --------------------

        # Create a dummy root
        xml = cStringIO.StringIO('<dummy>%s</dummy>' % source.read())
        source.close()

        root = ET.parse(xml, parser).getroot()
        for e in root[:]:
            # Attach the renderer to each roots
            e._renderer = self

        # Return the children of the dummy root
        return ([root.text] if root.text and not no_leading_text else []) + root[:]

    def parse_xmlstring(self, text, fragment=False, no_leading_text=False, **kw):
        """Parse a XML string

        In:
          - ``text`` -- can be a ``str`` or ``unicode`` string
          - ``fragment`` -- if ``True``, can parse a XML fragment i.e a XML without
            a unique root
          - ``no_leading_text`` -- if ``fragment`` is ``True``, ``no_leading_text``
            is ``False`` and the XML to parsed begins by a text, this text is keeped
          - ``kw`` -- keywords parameters are passed to the XML parser

        Return:
          - the root element of the parsed XML, if ``fragment`` is ``False``
          - a list of XML elements, if ``fragment`` is ``True``
        """
        if type(text) == unicode:
            text = text.encode(kw.setdefault('encoding', 'utf-8'))

        return self.parse_xml(cStringIO.StringIO(text), fragment, no_leading_text, **kw)


# ---------------------------------------------------------------------------

class Renderer(XmlRenderer):
    """The XML Renderer

    This renderer generate any tags you give

    .. code-block:: pycon

       >>> xml = xml.Renderer()
       >>> xml.foo(xml.bar).write_xmlstring()
       '<foo><bar/></foo>'
    """
    def __getattr__(self, name):
        """Any attribute access becomes a tag generation

        In:
          - ``name`` -- name of the tag to generate

        Return:
          - the generated tag
        """
        return self.makeelement(name)


# ---------------------------------------------------------------------------

if __name__ == '__main__':
    x = Renderer()
    x.namespaces = {'meld': 'http://www.plope.com/software/meld3'}

    with x.contacts:
        with x.contact.meld_id('contact'):
            x << x.name.meld_id('name')
            with x.address.meld_id('addr'):
                x << 'ici, rennes'

    print x.root.write_xmlstring(xml_declaration=True, pretty_print=True)
    print

    for (e, (name, addr)) in x.root.repeat((('bill', 'seatle'), ('steve', 'cupertino')), 'contact'):
        e.findmeld('name').text = name
        e.findmeld('addr').text = addr

    print x.root.write_xmlstring(pretty_print=True)
    print

    # -----------------------------------------------------------------------

    x = Renderer()
    x << x.foo
    x << x.bar

    print x.all(x.root).write_xmlstring(xml_declaration=True, pretty_print=True)
    print

    # -----------------------------------------------------------------------

    t1 = x.parse_xmlstring('''
    <a>avant<x>kjkjkl</x>#<b b='b'>hello</b>apres</a>
    ''')

    t2 = x.parse_xmlstring('''
    <toto>xxx<titi a='a'>world</titi>yyy</toto>
    ''')

    t1[1].replace(t2[0])
    # t1[1].replace('new')
    # t1[1].replace()

    print t1.write_xmlstring()



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Securiy API for the applications"""

from nagare import local, partial


# ---------------------------------------------------------------------------

# API to access the security context

def _get_user():
    """Return the current user

    Return:
      - the user object (created by the security manager)
    """
    return local.request.user


def get_user():
    """Return the current user

    Return:
      - the user object (created by the security manager) if not expired
    """
    user = _get_user()
    return user if user is not None and not user.expired else None


def set_user(user):
    """Change the user

    In:
      - ``user`` -- the current user
    """
    local.request.user = user


def get_manager():
    """Return the security manager

    Each application has a dedicated security manager

    Return:
      - the security manager
    """
    return local.request.security_manager


def set_manager(manager):
    """Change the security manager

    In:
      - ``manager`` -- the new security manager
    """
    local.request.security_manager = manager


# ---------------------------------------------------------------------------

# def flatten(*args):
#     return sum([flatten(*x) if hasattr(x, '__iter__') else (x,) for x in args], ())


def has_permissions(perm, subject=None):
    """Check that the current user has the permissions ``perm``
    on the object ``subject``

    Forward the call to the generic method ``has_permission()`` of the
    current security manager

    .. note::

      The default generic method can check a single permission or a list of
      permissions

    In:
      - ``perm`` -- permission(s)
      - ``subject`` -- object to check the permissions on

    Return:
      - True if the access is granted
      - Else a ``security.common.denial`` object
    """
    return get_manager().has_permission(get_user(), perm, subject)


def check_permissions(perm, subject=None):
    """Control that the current user has the permissions ``perm``
    on the object ``subject``

    Forward the call to the generic method ``has_permission()`` of the
    current security manager.

    Then let the security manager acts if the permission is denied.

    .. note::

      The default generic method can check a single permission or a list of
      permissions

    In:
      - ``perm`` -- permission(s)
      - ``subject`` -- object to check the permissions on

    Return:
      - True if the access is granted
      - Else a ``security.common.denial`` object
    """
    credential = has_permissions(perm, subject)
    if not credential:
        if credential is False:
            credential = ''
        get_manager().denies(credential)

    return credential


def call_with_permissions(self, __action, __perm, __subject, *args, **kw):
    """Call a function or method only if permit

    In:
      - ``self`` -- if ``None`` then ``__action`` is a function else a method
      - ``__action`` -- function or method to call
      - ``__perm`` -- permission(s) to check
      - ``__subject`` -- object to check the permissions on
      - ``args``, ``kw`` -- ``__action`` parameters

    Return:
      - ``__action`` return
    """
    check_permissions(__perm, __subject or self)
    return __action(self, *args, **kw) if self else __action(*args, **kw)


def wrapper(action, perm, subject):
    """Wrap a function or method into a wrapper that will check the user permissions

    In:
      - ``action`` -- function or method to wrapper
      - ``perm`` -- permission(s) to check
      - ``subject`` -- object to check the permissions on

    Return:
      - new action
    """
    if perm is not None:
        action = partial.Partial(call_with_permissions, None, action, perm, subject)

    return action


def permissions(perm, subject=None):
    """Decorator to check the permissions of the current user

    The ``subject`` will be the first argument of the decorated method

    In:
      - ``perm`` -- permission(s)
      - ``subject`` -- object to check the permissions on or the first argument
                       of the decorated method if ``None``
    """
    # perm = flatten(perm)
    return lambda f: partial.Decorator(f, call_with_permissions, perm, subject)


permissions_with_subject = permissions  # Obsolete



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Various tools used be the administrative commands"""

import os

import pkg_resources
import configobj

from nagare.admin import reference
from nagare import wsgi, config

# The default application configuration
# -------------------------------------

application_options_spec = {
    'application': dict(
        path='string()',  # Where to find the application object (see ``load_object()``) ?
        name='string()',  # URL for the application
        debug='boolean(default=False)',  # Debug web page activated ?

        redirect_after_post='boolean(default=False)',  # Follow the PRG pattern ?
        always_html='boolean(default=True)',  # Don't generate xhtml, even if it's a browser capability ?
        wsgi_pipe='string(default="")',  # Method to create the WSGI middlewares pipe
        static='string(default="$root/static")'  # Default directory of the static files
    ),

    'database': dict(
        activated='boolean(default=False)',  # Activate the database engine ?
        uri='string(default="")',  # Database connection string
        metadata='string(default="")',  # Database metadata : database entities description
        populate='string(default="")',  # Method to call after the database tables creation
        debug='boolean(default=False)',  # Set the database engine in debug mode ?
        __many__=dict(  # Database sub-sections
            activated='boolean(default=False)',
            populate='string(default="")'
        )
    ),

    'logging': dict()
}


def read_application_options(cfgfile, error, default={}):
    """Read the configuration file for the application

    In:
      - ``cfgfile`` -- path to an application configuration file
      - ``error`` -- the function to call in case of configuration errors
      - ``default`` -- optional default values

    Return:
      - a ``ConfigObj`` of the application parameters
    """
    if default:
        default['name'] = 'string(default="%s")' % configobj.ConfigObj(cfgfile)['application']['name']

    spec = configobj.ConfigObj(default)
    spec.merge(application_options_spec)

    choices = ', '. join(['"%s"' % entry.name for entry in pkg_resources.iter_entry_points('nagare.sessions')])
    spec.merge({'sessions': {'type': 'option(%s, default="")' % (choices + ', ""')}})

    conf = configobj.ConfigObj(cfgfile, configspec=spec, interpolation='Template' if default else None)
    config.validate(cfgfile, conf, error)

    # The database sub-sections inherit from the database section
    spec['database']['__many__'].merge(dict(
        uri='string(default=%s)' % str(conf['database']['uri']),
        metadata='string(default=%s)' % str(conf['database']['metadata']),
        debug='boolean(default=%s)' % str(conf['database']['debug']),
    ))
    conf = configobj.ConfigObj(cfgfile, configspec=spec, interpolation='Template' if default else None)
    config.validate(cfgfile, conf, error)

    if not conf['sessions']['type']:
        del conf['sessions']['type']

    return conf


def read_application(cfgfile, error):
    """Read the configuration file for the application and create the application object

    In:
      - ``cfgfile`` -- name of a registered application or path to an application configuration file
      - ``error`` -- the function to call in case of configuration errors

    Return:
      - a tuple:

        - name of the application configuration file
        - the application object
        - the distribution of the application
        - a ``ConfigObj`` of the application parameters
    """
    if not os.path.isfile(cfgfile):
        # The name of a registered application is given, find its configuration file

        # Get all the registered applications
        apps = dict([(entry.name, entry) for entry in pkg_resources.iter_entry_points('nagare.applications')])

        # Get the application
        app = apps.get(cfgfile)
        if app is None:
            error("application '%s' not found (use 'nagare-admin serve' to see the available applications)" % cfgfile)

        # From the directory of the application, get its configuration file
        requirement = pkg_resources.Requirement.parse(app.dist.project_name)
        cfgfile = pkg_resources.resource_filename(requirement, os.path.join('conf', cfgfile + '.cfg'))

    # Read the application configuration file
    aconf = read_application_options(cfgfile, error)

    # From the path of the application, create the application object
    (app, dist) = reference.load_object(aconf['application']['path'])

    here = os.path.abspath(os.path.dirname(cfgfile))
    root = os.path.abspath(os.path.join(here, '..')) if dist is None else dist.location

    defaults = {
        'here': 'string(default="%s")' % here,
        'root': 'string(default="%s")' % root
    }

    # Re-read the application configuration, with some substitution variables
    aconf = read_application_options(cfgfile, error, defaults)

    return cfgfile, app, dist, aconf


# ---------------------------------------------------------------------------

def get_database(conf, debug):
    """Read the database settings

    The location of the metadata object is read from the configuration file

    In:
      - ``conf`` -- the ``ConfigObj`` object, created from the configuration file
      - ``debug`` -- debug mode for the database engine

    Return:
      - the tuple:
        - metadata object
        - database uri
        - database debug mode
        - database engines settings
    """
    metadata = conf.get('metadata')

    if not conf['activated'] or not metadata:
        return None

    # Import the metadata object
    metadata = reference.load_object(metadata)[0]

    # All the parameters, of the [database] section, with an unknown name are
    # given to the database engine
    engine_conf = dict([(k, v) for (k, v) in conf.items() if k not in ('uri', 'activated', 'metadata', 'debug', 'populate')])

    return metadata, conf['uri'], debug, engine_conf


def activate_WSGIApp(
    app,
    cfgfile, aconf, error,
    project_name='',
    static_path=None, static_url=None,
    data_path=None,
    publisher=None, sessions_manager=None,
    debug=False
):
    """Set all the properties of a WSGIApp application

    In:
      - ``app`` -- the WSGIApp application or the application root object factory
      - ``cfgfile`` -- the path to the configuration file
      - ``aconf`` -- the ``ConfigObj`` object, created from the configuration file
      - ``error`` -- the function to call in case of configuration errors
      - ``project_name`` -- name of the distutils distribution where the app is located
      - ``static_path`` -- the directory where the static contents of the application
        are located
      - ``static_url`` -- the url of the static contents of the application
      - ``data_path`` -- the directory where the data of the application are located
      - ``publisher`` -- the publisher of the application
      - ``session_manager`` -- the sessions manager
      - ``debug`` -- flag to display the generated SQL statements

    Return:
      - a tuple:
          - the ``wsgi.WSGIApp`` object
          - tuples (application databases settings, application databases populate functions)
    """
    databases = []
    populates = []
    # Get all the databases settings
    for (section, content) in aconf['database'].items():
        if isinstance(content, configobj.Section):
            database = get_database(content, content['debug'] or debug)
            if database:
                databases.append(database)
                populates.append(content['populate'])
            del aconf['database'][section]

    database = get_database(aconf['database'], aconf['database']['debug'] or debug)
    if database:
        databases.append(database)
        populates.append(aconf['database']['populate'])

    app = wsgi.create_WSGIApp(app)

    app.set_config(cfgfile, aconf, error)

    if static_path is not None:
        app.set_static_path(static_path)

    if static_url is not None:
        app.set_static_url(static_url)

    if data_path is not None:
        app.set_data_path(data_path)

    if publisher:
        app.set_publisher(publisher)

    if sessions_manager:
        app.set_sessions_manager(sessions_manager)

    if databases:
        app.set_databases(databases)

    if project_name:
        app.set_project(project_name)

    return app, zip(databases, populates)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""The FAPWS3__  publisher

__ http://william-os4y.livejournal.com/
"""

import fapws.base
import fapws._evwsgi as evwsgi

from nagare.publishers import common


class Publisher(common.Publisher):
    """The FAPWS3 publisher"""

    # Possible command line options with the default values
    # ------------------------------------------------------

    spec = dict(host='string(default=None)', port='integer(default=None)')

    def serve(self, filename, conf, error):
        """Run the publisher

        In:
          -  ``filename`` -- the path to the configuration file
          - ``conf`` -- the ``ConfigObj`` object, created from the configuration file
          - ``error`` -- the function to call in case of configuration errors
        """
        (host, port, conf) = self._validate_conf(filename, conf, error)

        # The publisher is an events based server so call once the ``on_new_process()`` method
        self.on_new_process()

        evwsgi.start(host, str(port))
        evwsgi.set_base_module(fapws.base)
        evwsgi.wsgi_cb(self.urls)
        evwsgi.run()



# Encoding: utf-8

# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

import unittest

from lxml import etree

from nagare.namespaces import xml, xhtml
from nagare.serializer import serialize


class TestSerializer(unittest.TestCase):
    def test_html(self):
        h = xhtml.Renderer()

        r = serialize(h.p('hello'), 'text/html', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/html', '<p>hello</p>\n'))
        r = serialize(h.p(u'été'), 'text/html', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/html', '<p>\xc3\xa9t\xc3\xa9</p>\n'))
        r = serialize(h.p('hello'), 'text/html', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/html', '<!DOCTYPE html>\n<p>hello</p>\n'))

    def test_xhtml(self):
        h = xhtml.Renderer()

        r = serialize(h.p('hello'), 'application/xhtml+xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('application/xhtml+xml', '<p>hello</p>'))
        r = serialize(h.p(u'été'), 'application/xhtml+xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('application/xhtml+xml', '<p>\xc3\xa9t\xc3\xa9</p>'))
        r = serialize(h.p('hello'), 'application/xhtml+xml', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('application/xhtml+xml', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<p>hello</p>"))

    def test_xml(self):
        x = xml.Renderer()

        r = serialize(x.person('hello'), 'text/xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/xml', '<person>hello</person>'))
        r = serialize(x.person(u'été'), 'text/xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/xml', '<person>\xc3\xa9t\xc3\xa9</person>'))
        r = serialize(x.person('hello'), 'text/xml', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/xml', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<person>hello</person>"))

    def test_comment(self):
        x = xml.Renderer()

        r = serialize(x.comment('hello'), 'text/xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/xml', '<!--hello-->'))
        r = serialize(x.comment(u'été'), 'text/xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/xml', '<!--\xc3\xa9t\xc3\xa9-->'))
        r = serialize(x.comment('hello'), 'text/xml', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/xml', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<!--hello-->"))

    def test_pi(self):
        x = xml.Renderer()

        r = serialize(x.processing_instruction('hello'), 'text/xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/xml', '<?hello ?>'))
        r = serialize(x.processing_instruction(u'été'), 'text/xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/xml', '<?\xc3\xa9t\xc3\xa9 ?>'))
        r = serialize(x.processing_instruction('hello'), 'text/xml', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/xml', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<?hello ?>"))

    def test_etree(self):
        e = etree.Element('person')
        e.text = 'hello'
        r = serialize(e, 'text/xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/xml', '<person>hello</person>'))
        e = etree.Element('person')
        e.text = u'été'
        r = serialize(e, 'text/xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/xml', '<person>\xc3\xa9t\xc3\xa9</person>'))
        e = etree.Element('person')
        e.text = 'hello'
        r = serialize(e, 'text/xml', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/xml', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<person>hello</person>"))

    def test_str(self):
        r = serialize('hello world', 'text/msg', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/msg', 'hello world'))
        r = serialize(u'été'.encode('utf-8'), 'text/msg', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/msg', '\xc3\xa9t\xc3\xa9'))
        r = serialize('hello world', 'text/msg', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/msg', 'hello world'))
        r = serialize('hello world', '', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/plain', 'hello world'))
        r = serialize(u'été'.encode('utf-8'), '', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/plain', '\xc3\xa9t\xc3\xa9'))
        r = serialize('hello world', '', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/plain', 'hello world'))

    def test_unicode(self):
        r = serialize(u'hello world', 'text/msg', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/msg', 'hello world'))
        r = serialize(u'été', 'text/msg', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/msg', '\xc3\xa9t\xc3\xa9'))
        r = serialize(u'hello world', 'text/msg', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/msg', 'hello world'))
        r = serialize(u'hello world', '', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/plain', 'hello world'))
        r = serialize(u'été', '', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/plain', '\xc3\xa9t\xc3\xa9'))
        r = serialize(u'hello world', '', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/plain', 'hello world'))

    def test_list(self):
        h = xhtml.Renderer()
        x = xml.Renderer()

        e = etree.Element('person')
        e.text = 'hello'

        l = [h.p('hello'), x.person('hello'), x.comment('hello'), x.processing_instruction('hello'), e, 'hello', u'hello']

        r = serialize(l, 'text/html', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/html', '<p>hello</p>\n<person>hello</person>\n<!--hello-->\n<?hello >\n<person>hello</person>\nhellohello'))
        r = serialize(l, 'text/html', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/html', '<!DOCTYPE html>\n<p>hello</p>\n<person>hello</person>\n<!--hello-->\n<?hello >\n<person>hello</person>\nhellohello'))

        r = serialize(l, 'application/xhtml+xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('application/xhtml+xml', '<p>hello</p><person>hello</person><!--hello--><?hello ?><person>hello</person>hellohello'))
        r = serialize(l, 'application/xhtml+xml', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('application/xhtml+xml', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<p>hello</p><person>hello</person><!--hello--><?hello ?><person>hello</person>hellohello"))

        r = serialize(l, 'text/msg', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/msg', '<p>hello</p>\n<person>hello</person><!--hello--><?hello ?><person>hello</person>hellohello'))
        r = serialize(l, 'text/msg', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/msg', '<!DOCTYPE html>\n<p>hello</p>\n<person>hello</person><!--hello--><?hello ?><person>hello</person>hellohello'))

        r = serialize(l, '', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('', '<p>hello</p>\n<person>hello</person><!--hello--><?hello ?><person>hello</person>hellohello'))
        r = serialize(l, '', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('', '<!DOCTYPE html>\n<p>hello</p>\n<person>hello</person><!--hello--><?hello ?><person>hello</person>hellohello'))

        l = l[1:] + [l[0]]
        r = serialize(l, 'text/html', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/html', '<person>hello</person>\n<!--hello-->\n<?hello >\n<person>hello</person>\nhellohello<p>hello</p>\n'))
        r = serialize(l, 'text/html', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/html', '<!DOCTYPE html>\n<person>hello</person>\n<!--hello-->\n<?hello >\n<person>hello</person>\nhellohello<p>hello</p>\n'))

        r = serialize(l, 'application/xhtml+xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('application/xhtml+xml', '<person>hello</person><!--hello--><?hello ?><person>hello</person>hellohello<p>hello</p>'))
        r = serialize(l, 'application/xhtml+xml', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('application/xhtml+xml', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<person>hello</person><!--hello--><?hello ?><person>hello</person>hellohello<p>hello</p>"))

        r = serialize(l, 'text/msg', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/msg', "<person>hello</person><!--hello--><?hello ?><person>hello</person>hellohello<p>hello</p>\n"))
        r = serialize(l, 'text/msg', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/msg', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<person>hello</person><!--hello--><?hello ?><person>hello</person>hellohello<p>hello</p>\n"))

        r = serialize(l, '', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('', '<person>hello</person><!--hello--><?hello ?><person>hello</person>hellohello<p>hello</p>\n'))
        r = serialize(l, '', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<person>hello</person><!--hello--><?hello ?><person>hello</person>hellohello<p>hello</p>\n"))

        l = l[1:] + [l[0]]
        r = serialize(l, 'text/html', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/html', '<!--hello-->\n<?hello >\n<person>hello</person>\nhellohello<p>hello</p>\n<person>hello</person>\n'))
        r = serialize(l, 'text/html', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/html', '<!DOCTYPE html>\n<!--hello-->\n<?hello >\n<person>hello</person>\nhellohello<p>hello</p>\n<person>hello</person>\n'))

        r = serialize(l, 'application/xhtml+xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('application/xhtml+xml', '<!--hello--><?hello ?><person>hello</person>hellohello<p>hello</p><person>hello</person>'))
        r = serialize(l, 'application/xhtml+xml', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('application/xhtml+xml', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<!--hello--><?hello ?><person>hello</person>hellohello<p>hello</p><person>hello</person>"))

        r = serialize(l, 'text/msg', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/msg', '<!--hello--><?hello ?><person>hello</person>hellohello<p>hello</p>\n<person>hello</person>'))
        r = serialize(l, 'text/msg', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/msg', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<!--hello--><?hello ?><person>hello</person>hellohello<p>hello</p>\n<person>hello</person>"))

        r = serialize(l, '', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('', '<!--hello--><?hello ?><person>hello</person>hellohello<p>hello</p>\n<person>hello</person>'))
        r = serialize(l, '', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<!--hello--><?hello ?><person>hello</person>hellohello<p>hello</p>\n<person>hello</person>"))

        l = l[1:] + [l[0]]
        r = serialize(l, 'text/html', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/html', '<?hello >\n<person>hello</person>\nhellohello<p>hello</p>\n<person>hello</person>\n<!--hello-->\n'))
        r = serialize(l, 'text/html', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/html', '<!DOCTYPE html>\n<?hello >\n<person>hello</person>\nhellohello<p>hello</p>\n<person>hello</person>\n<!--hello-->\n'))

        r = serialize(l, 'application/xhtml+xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('application/xhtml+xml', '<?hello ?><person>hello</person>hellohello<p>hello</p><person>hello</person><!--hello-->'))
        r = serialize(l, 'application/xhtml+xml', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('application/xhtml+xml', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<?hello ?><person>hello</person>hellohello<p>hello</p><person>hello</person><!--hello-->"))

        r = serialize(l, 'text/msg', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/msg', '<?hello ?><person>hello</person>hellohello<p>hello</p>\n<person>hello</person><!--hello-->'))
        r = serialize(l, 'text/msg', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/msg', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<?hello ?><person>hello</person>hellohello<p>hello</p>\n<person>hello</person><!--hello-->"))

        r = serialize(l, '', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('', '<?hello ?><person>hello</person>hellohello<p>hello</p>\n<person>hello</person><!--hello-->'))
        r = serialize(l, '', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<?hello ?><person>hello</person>hellohello<p>hello</p>\n<person>hello</person><!--hello-->"))

        l = l[1:] + [l[0]]
        r = serialize(l, 'text/html', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/html', '<person>hello</person>\nhellohello<p>hello</p>\n<person>hello</person>\n<!--hello-->\n<?hello >\n'))
        r = serialize(l, 'text/html', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/html', '<!DOCTYPE html>\n<person>hello</person>\nhellohello<p>hello</p>\n<person>hello</person>\n<!--hello-->\n<?hello >\n'))

        r = serialize(l, 'application/xhtml+xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('application/xhtml+xml', '<person>hello</person>hellohello<p>hello</p><person>hello</person><!--hello--><?hello ?>'))
        r = serialize(l, 'application/xhtml+xml', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('application/xhtml+xml', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<person>hello</person>hellohello<p>hello</p><person>hello</person><!--hello--><?hello ?>"))

        r = serialize(l, 'text/msg', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/msg', '<person>hello</person>hellohello<p>hello</p>\n<person>hello</person><!--hello--><?hello ?>'))
        r = serialize(l, 'text/msg', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/msg', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<person>hello</person>hellohello<p>hello</p>\n<person>hello</person><!--hello--><?hello ?>"))

        r = serialize(l, '', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('', '<person>hello</person>hellohello<p>hello</p>\n<person>hello</person><!--hello--><?hello ?>'))
        r = serialize(l, '', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('', "<?xml version='1.0' encoding='utf-8'?>\n<!DOCTYPE html>\n<person>hello</person>hellohello<p>hello</p>\n<person>hello</person><!--hello--><?hello ?>"))

        l = l[1:] + [l[0]]
        r = serialize(l, 'text/html', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/html', 'hellohello<p>hello</p>\n<person>hello</person>\n<!--hello-->\n<?hello >\n<person>hello</person>\n'))
        r = serialize(l, 'text/html', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/html', 'hellohello<p>hello</p>\n<person>hello</person>\n<!--hello-->\n<?hello >\n<person>hello</person>\n'))

        r = serialize(l, 'application/xhtml+xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('application/xhtml+xml', 'hellohello<p>hello</p><person>hello</person><!--hello--><?hello ?><person>hello</person>'))
        r = serialize(l, 'application/xhtml+xml', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('application/xhtml+xml', 'hellohello<p>hello</p><person>hello</person><!--hello--><?hello ?><person>hello</person>'))

        r = serialize(l, 'text/msg', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/msg', 'hellohello<p>hello</p>\n<person>hello</person><!--hello--><?hello ?><person>hello</person>'))
        r = serialize(l, 'text/msg', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/msg', 'hellohello<p>hello</p>\n<person>hello</person><!--hello--><?hello ?><person>hello</person>'))

        r = serialize(l, '', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('', 'hellohello<p>hello</p>\n<person>hello</person><!--hello--><?hello ?><person>hello</person>'))
        r = serialize(l, '', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('', 'hellohello<p>hello</p>\n<person>hello</person><!--hello--><?hello ?><person>hello</person>'))

        l = l[1:] + [l[0]]
        r = serialize(l, 'text/html', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/html', 'hello<p>hello</p>\n<person>hello</person>\n<!--hello-->\n<?hello >\n<person>hello</person>\nhello'))
        r = serialize(l, 'text/html', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/html', 'hello<p>hello</p>\n<person>hello</person>\n<!--hello-->\n<?hello >\n<person>hello</person>\nhello'))

        r = serialize(l, 'application/xhtml+xml', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('application/xhtml+xml', 'hello<p>hello</p><person>hello</person><!--hello--><?hello ?><person>hello</person>hello'))
        r = serialize(l, 'application/xhtml+xml', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('application/xhtml+xml', 'hello<p>hello</p><person>hello</person><!--hello--><?hello ?><person>hello</person>hello'))

        r = serialize(l, 'text/msg', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('text/msg', 'hello<p>hello</p>\n<person>hello</person><!--hello--><?hello ?><person>hello</person>hello'))
        r = serialize(l, 'text/msg', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('text/msg', 'hello<p>hello</p>\n<person>hello</person><!--hello--><?hello ?><person>hello</person>hello'))

        r = serialize(l, '', '<!DOCTYPE html>', False)
        self.assertEqual(r, ('', 'hello<p>hello</p>\n<person>hello</person><!--hello--><?hello ?><person>hello</person>hello'))
        r = serialize(l, '', '<!DOCTYPE html>', True)
        self.assertEqual(r, ('', 'hello<p>hello</p>\n<person>hello</person><!--hello--><?hello ?><person>hello</person>hello'))



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Various pre-defined components"""

from nagare import presentation, var


class Ask(object):
    """Ask the user to enter a line of text

    The text entered is answered
    """
    def __init__(self, msg):
        """Initialization

        In:
          - ``msg`` -- message to display
        """
        self.msg = msg

    @classmethod
    def answer(cls, comp, r):
        comp.answer(r())


@presentation.render_for(Ask)
def render(self, h, comp, *args):
    """The view is a simple form with a text input control

    In:
      - ``h`` -- the renderer
      - ``comp`` -- the component
      - ``model`` -- the name of the view

    Return:
      - a tree
    """
    r = var.Var()

    return h.form(
        self.msg, ' ',
        h.input.action(r),
        h.input(type='submit', value='Send').action(self.answer, comp, r)
    )


class Confirm(object):
    """Display a confirmation message
    """

    def __init__(self, msg):
        """Initialization

        In:
          - ``msg`` -- message to display
        """
        self.msg = msg


@presentation.render_for(Confirm)
def render_confirm(self, h, comp, *args):
    """The view is a simple form with the text and a submit button

    In:
      - ``h`` -- the renderer
      - ``comp`` -- the component
      - ``model`` -- the name of the view

    Return:
      - a tree
    """
    return h.form(
        self.msg, h.br,
        h.input(type='submit', value='ok').action(comp.answer)
    )


class View(object):
    """A differed view

    Can be useful to add a view for a ``component.Task``
    """
    def __init__(self, view):
        """Initialization

        In:
          - ``view`` -- a function that will be called to render this object
        """
        self.view = view


@presentation.render_for(View)
def render_view(self, h, *args):
    """To render this object, call the differed view function

    In:
      - ``h`` -- the renderer

    Return:
      - a tree
    """
    return self.view(h)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""This is the admin application. By default, when launched, it's mapped to url
``/admin`` and ``/``.

The default view of this application is the administration view
of the framework.
"""

from __future__ import with_statement

import operator

import pkg_resources
import configobj

from nagare import component, presentation, wsgi, config


class Admin(object):
    """The root component of the admin application

    It loads all the components registered under the ``nagare.admin`` entry point and renders them
    into its view.
    """
    def __init__(self, apps):
        """Load the components registered under the ``nagare.admin`` entry point

        In:
          - ``apps`` -- list of all the applications launched
        """
        self.apps = [(app_name, app_urls) for (app, app_name, app_urls) in apps]

        # Load the admin objects
        components = [entry.load()(apps) for entry in pkg_resources.iter_entry_points('nagare.admin')]

        # Sort them according to their ``priority`` attribute
        components.sort(key=operator.attrgetter('priority'))

        # Create the components
        self.components = map(component.Component, components)


@presentation.render_for(Admin)
def render(self, h, comp, *args):
    """Aggregates all the default views of the components"""

    h.head.css_url('/static/nagare/application.css')
    h.head << h.head.title('Nagare Administration interface')

    h.head << h.head.style('''
        #content { margin-left: 0 !important }
        #footer { text-align: right }
        #footer a { padding-right: 10px }
    ''')

    with h.div(id='body'):
        h << h.a(h.img(src='/static/nagare/img/logo_small.png'), id='logo', href='http://www.nagare.org/', title='Nagare home')

        with h.div(id='content'):
            h << h.div('Administration interface', id='title')

            with h.div(id='main'):
                h << [c.on_answer(comp.call) for c in self.components]

    with h.div(' ', id='footer'):
        h << h.a(u'\N{Copyright Sign} Net-ng', href='http://www.net-ng.com')

    return h.root


# ---------------------------------------------------------------------------

application_spec = {'application': {'as_root': 'boolean(default=True)'}}


class WSGIApp(wsgi.WSGIApp):
    """The admin application

    Read the additional boolean parameter ``as_root`` into the ``[application]`` section.
    If ``True``, mappes itself to the ``/`` url.
    """

    def set_config(self, config_filename, conf, error):
        """Read the value of the ``as_root`` parameter and keeps the list of all
        the launched applications

        In:
          - ``config_filename`` -- the path to the configuration file
          - ``config`` -- the ``ConfigObj`` object, created from the configuration file
          - ``error`` -- the function to call in case of configuration errors
        """
        conf = configobj.ConfigObj(conf, configspec=configobj.ConfigObj(application_spec))
        config.validate(config_filename, conf, error)

        self.as_root = conf['application']['as_root']
        self.application_path = conf['application']['path']

        super(WSGIApp, self).set_config(config_filename, conf, error)

    def set_publisher(self, publisher):
        """Read the value of the ``as_root`` parameter and keeps a reference to the publisher

        In:
          - ``publisher`` -- the publisher of the application
        """
        super(WSGIApp, self).set_publisher(publisher)

        self.publisher = publisher

        if self.as_root:
            publisher.register_application(self.application_path, '', self, self)

    def create_root(self):
        """
        Create an ``admin`` object

        Return:
          - the admin object
        """
        # Create the ``admin`` object with the list of all the launched application
        return self.root_factory(self.publisher.get_registered_applications())


# ---------------------------------------------------------------------------

# The admin application singleton.
# The given factory creates a component from a ``admin`` object
app = WSGIApp(lambda apps: component.Component(Admin(apps)))



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Empty security manager"""

from nagare.security import common


class Manager(common.Authentication, common.Rules):
    """A security manager is typically a mix-in of an authentication
    manager and security rules
    """
    pass



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Sessions managed in memory

These sessions managers keep:
  - the last recently used ``DEFAULT_NB_SESSIONS`` sessions
  - for each session, the last recently used ``DEFAULT_NB_STATES`` states
"""

from nagare import local
from nagare.sessions import ExpirationError, common, lru_dict
from nagare.sessions.serializer import Pickle

DEFAULT_NB_SESSIONS = 10000
DEFAULT_NB_STATES = 20


class Sessions(common.Sessions):
    """Sessions manager for states kept in memory
    """
    spec = dict(
        common.Sessions.spec,
        nb_sessions='integer(default=%d)' % DEFAULT_NB_SESSIONS,
        nb_states='integer(default=%d)' % DEFAULT_NB_STATES
    )

    def __init__(self, nb_sessions=DEFAULT_NB_SESSIONS, nb_states=DEFAULT_NB_STATES, **kw):
        """Initialization

        In:
          - ``nb_sessions`` -- maximum number of sessions kept in memory
          - ``nb_states`` -- maximum number of states, for each sessions, kept in memory
        """
        super(Sessions, self).__init__(**kw)

        self.nb_states = nb_states
        self._sessions = lru_dict.ThreadSafeLRUDict(nb_sessions)

    def set_config(self, filename, conf, error):
        """Read the configuration parameters

        In:
          - ``filename`` -- path to the configuration file
          - ``conf`` -- ``ConfigObj`` object created from the configuration file
          - ``error`` -- function to call in case of configuration errors
        """
        # Let's the super class validate the configuration file
        conf = super(Sessions, self).set_config(filename, conf, error)

        self.nb_states = conf['nb_states']
        self._sessions = lru_dict.ThreadSafeLRUDict(conf['nb_sessions'])

        return conf

    def check_session_id(self, session_id):
        """Test if a session exist

        In:
          - ``session_id`` -- id of a session

        Return:
          - is ``session_id`` the id of an existing session?
        """
        return session_id in self._sessions

    def create_lock(self, session_id):
        """Create a new lock for a session

        In:
          - ``session_id`` -- session id

        Return:
          - the lock
        """
        return local.worker.create_lock()

    def get_lock(self, session_id):
        """Retrieve the lock of a session

        In:
          - ``session_id`` -- session id

        Return:
          - the lock
        """
        try:
            return self._sessions[session_id][1]
        except KeyError:
            raise ExpirationError()

    def create(self, session_id, secure_id, lock):
        """Create a new session

        In:
          - ``session_id`` -- id of the session
          - ``secure_id`` -- the secure number associated to the session
          - ``lock`` -- the lock of the session
        """
        self._sessions[session_id] = [0, lock, secure_id, None, lru_dict.LRUDict(self.nb_states)]

    def delete(self, session_id):
        """Delete a session

        In:
          - ``session_id`` -- id of the session to delete
        """
        del self._sessions[session_id]

    def fetch_state(self, session_id, state_id):
        """Retrieve a state with its associated objects graph

        In:
          - ``session_id`` -- session id of this state
          - ``state_id`` -- id of this state

        Return:
          - id of the latest state
          - secure number associated to the session
          - data kept into the session
          - data kept into the state
        """
        try:
            last_state_id, _, secure_id, session_data, states = self._sessions[session_id]
            state_data = states[state_id]
        except KeyError:
            raise ExpirationError()

        return last_state_id, secure_id, session_data, state_data

    def store_state(self, session_id, state_id, secure_id, use_same_state, session_data, state_data):
        """Store a state and its associated objects graph

        In:
          - ``session_id`` -- session id of this state
          - ``state_id`` -- id of this state
          - ``secure_id`` -- the secure number associated to the session
          - ``use_same_state`` -- is this state to be stored in the previous snapshot?
          - ``session_data`` -- data to keep into the session
          - ``state_data`` -- data to keep into the state
        """
        session = self._sessions[session_id]

        if not use_same_state:
            session[0] += 1

        session[3] = session_data
        session[4][state_id] = state_data


class SessionsWithPickledStates(Sessions):
    """Sessions manager for states pickled / unpickled in memory
    """
    spec = Sessions.spec.copy()
    spec['serializer'] = 'string(default="nagare.sessions.serializer:Pickle")'

    def __init__(self, serializer=None, **kw):
        """Initialization

        In:
          - ``serializer`` -- serializer / deserializer of the states
        """
        super(SessionsWithPickledStates, self).__init__(serializer=serializer or Pickle, **kw)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

from __future__ import with_statement

import os

from nagare.namespaces import xsl
from nagare.namespaces import xml
from nagare.namespaces import xhtml


def test1():
    """ XSL namespace unit test - simple xsl transformation 1 """
    x = xsl.Renderer()
    x.namespaces = {'xsl': 'http://www.w3.org/1999/XSL/Transform'}
    x.default_namespace = 'xsl'

    styleSheet = x.stylesheet(
        x.output(encoding="utf-8", method="html"),
        x.template(x.copy(x.apply_templates(select="@*|node()")), match="@*|node()"),
        x.template(x.element(name="head"),
                   x.element(x.element(x.apply_templates(select="@*|node()"), name="body"), name="html"),
                   match="helloWorlds"),
        x.template(x.element(x.value_of(select="@language"), ' : ', x.value_of(select="."), name="h1"), match="helloWorld")
    )

    x = xml.Renderer()

    f = open(os.path.join(os.path.dirname(__file__), 'test_xmlns_1.xml'))
    root = x.parse_xmlstring(f.read())
    f.close()

    r = root.getroottree().xslt(styleSheet)

    f = open(os.path.join(os.path.dirname(__file__), 'helloworld.html'))
    xmlToCompare = f.read()
    f.close()

    assert r.__str__().strip() == xmlToCompare


def test2():
    """ XSL namespace unit test - simple xsl transformation 2 """
    x = xsl.Renderer()
    x.namespaces = {'xsl': 'http://www.w3.org/1999/XSL/Transform'}
    x.default_namespace = 'xsl'

    styleSheet = x.stylesheet(
        x.output(encoding="utf-8"),
        x.template(
            x.copy_of(select="."), match="/"
        )
    )

    h = xhtml.Renderer()

    page = h.html(h.h1('Hello'), h.h2('World'))

    r = page.getroottree().xslt(styleSheet)

    assert r.xpath('//html/h1')[0].text == 'Hello'
    assert r.xpath('//html/h2')[0].text == 'World'



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Authentication manager for the digest HTTP authentication scheme"""

import time
import hashlib

import webob

from nagare.security import common


class Authentication(common.Authentication):
    """Authentication manager for the digest HTTP authentication scheme
    """
    def __init__(self, realm, private_key):
        """Initialization

        In:
          - ``realm`` -- authentication realm
          - ``private_key`` -- will be hashed to generate the digest
        """
        self.realm = str(realm)
        self.private_key = private_key

    def get_ids(self, request, response):
        """Return the data associated with the connected user

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object

        Return:
          - A tuple with the id of the user and all the challenge response parameters
        """
        authorization = request.headers.get('authorization')
        if authorization is not None:
            (scheme, data) = authorization.split(' ', 1)
            if scheme == 'Digest':
                data = [x.split('=', 1) for x in data.split(',')]
                data = dict([(k.lstrip(), v.strip('"')) for (k, v) in data])

                encoding = request.accept_charset.best_match(['iso-8859-1', 'utf-8'])
                data['encoding'] = encoding
                data['http_method'] = request.method

                return data.pop('username').decode(encoding), data

        return None, {'response': None, 'encoding': None}

    def check_password(
        self,
        username, password,
        response, encoding, realm='', uri='',
        nonce='', nc='', cnonce='', qop='', http_method='',
        **kw
    ):
        """Authentication

        In:
          - ``username`` -- user id
          - ``password`` -- real password of the user
          - ``encoding`` -- encoding of username and password on the client
          - ``response``, ``realm``, ``uri``, ``nonce``, ``nc``, ``cnonce``,
            ``qop`` -- elements of the challenge response

        Return:
          - a boolean
        """
        if (username is None) and (response is None):
            # Anonymous user
            return None

        # Make our side hash
        hda1 = hashlib.md5('%s:%s:%s' % (username.encode(encoding), realm, password.encode(encoding))).hexdigest()
        hda2 = hashlib.md5(http_method + ':' + uri).hexdigest()

        sig = '%s:%s:%s:%s:%s:%s' % (hda1, nonce, nc, cnonce, qop, hda2)

        # Compare our hash with the response
        return hashlib.md5(sig).hexdigest() == response

    def denies(self, detail):
        """Method called when a permission is denied

        In:
          - ``details`` -- a ``security.common.denial`` object
        """
        nonce = hashlib.md5('%s:%s' % (str(time.time()), self.private_key)).hexdigest()

        raise webob.exc.HTTPUnauthorized(
            str(detail),
            [('WWW-Authenticate', 'Digest realm="%s", nonce="%s", qop="auth"' % (self.realm, nonce))]
        )



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""The ``create-db`` and ``drop-db`` administrative commands

Create or delete the database tables of an application
"""

from __future__ import with_statement

import pkg_resources

from nagare import local, log, database
from nagare.admin import reference, util, command


def read_options(debug, args, error):
    """Activate all the database metadata objects of an application

    Several metadata objects can be activated if there are sub-sections into
    the [database] section.

    In:
      - ``debug`` -- flag to display the generated SQL statements
      - ``args`` -- arguments in the command lines: application to activate
      - ``error`` -- the function to call in case of configuration errors

    Return:
      - tuples (metadata object, populate function)
   """
    # If no application name is given, display the list of the registered applications
    if len(args) == 0:
        print 'Available applications:'
        app_names = [entry.name for entry in pkg_resources.iter_entry_points('nagare.applications')]
        for app_name in sorted(app_names):
            print ' -', app_name
        return ()

    if len(args) != 1:
        error('Bad number of parameters')

    # Read the configuration of the application
    (cfgfile, app, dist, aconf) = util.read_application(args[0], error)

    # Configure the local service
    local.worker = local.Process()
    local.request = local.Process()

    # Configure the logging service
    log.configure(aconf['logging'].dict(), aconf['application']['name'])
    log.activate()
    log.set_logger('nagare.application.' + aconf['application']['name'])

    return util.activate_WSGIApp(app, cfgfile, aconf, error, debug=debug)[1]


def create(parser, options, args):
    """Create the database tables of the application

    If the ``--drop`` option is on, delete the existing tables before to re-create them

    If the ``--no-populate`` option is off, call the populate function (if it exists)
    after the creation of the tables

    In:
      - ``parser`` -- the optparse.OptParser object used to parse the configuration file
      - ``options`` -- options in the command lines
      - ``args`` -- arguments in the command lines : application name
    """
    for (database_settings, populate) in read_options(options.debug, args, parser.error):
        database.set_metadata(*database_settings)

        with database.session.begin():
            if options.drop:
                database_settings[0].drop_all()

            database_settings[0].create_all()

            if options.populate and populate:
                reference.load_object(populate)[0]()


def drop(parser, options, args):
    """Delete the database tables of the application

    In:
      - ``parser`` -- the optparse.OptParser object used to parse the configuration file
      - ``options`` -- options in the command lines
      - ``args`` -- arguments in the command lines : application name
    """
    for (database_settings, populate) in read_options(options.debug, args, parser.error):
        database.set_metadata(*database_settings)

        with database.session.begin():
            database_settings[0].drop_all()


# ---------------------------------------------------------------------------

class DBCreate(command.Command):
    desc = 'Create the database of an application'

    @staticmethod
    def set_options(optparser):
        optparser.usage += ' <application>'

        optparser.add_option('-d', '--debug', action='store_const', const=True, default=False, dest='debug', help='debug mode for the database engine')
        optparser.add_option('--drop', action='store_const', const=True, default=False, dest='drop', help='drop the database tables before to re-create them')
        optparser.add_option('--no-populate', action='store_const', const=False, default=True, dest='populate', help='populate the database tables')

    run = staticmethod(create)


class DBDrop(command.Command):
    desc = 'Drop the database of an application'

    @staticmethod
    def set_options(optparser):
        optparser.usage += ' <application>'

        optparser.add_option('-d', '--debug', action='store_const', const=True, default=False, dest='debug', help='debug mode for the database engine')

    run = staticmethod(drop)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Scoped objects

Currently:

  - objects scoped to a worker, a handler of a request
  - objects scoped to a request (i.e a scoped cleared on each new request)
"""

import threading


class Thread(threading.local):
    """Objects with attributs scoped to the current thread
    """
    def clear(self):
        self.__dict__.clear()

    def create_lock(self):
        return threading.Lock()


class DummyLock(object):
    acquire = release = lambda self: None


class Process(object):
    """Objects with attributs scoped to the current thread
    """
    def clear(self):
        self.__dict__.clear()

    def create_lock(self):
        return DummyLock()


# ----------------------------------------------------------------------------

worker = None
request = None



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Applications administration view
"""

from __future__ import with_statement

from nagare import presentation


class Admin(object):
    priority = 100        # Order of the default view, into the administrative interface

    def __init__(self, apps):
        """Initialization

        In:
          - ``apps`` -- list of tuples (application, application name, application urls)
        """
        self.apps = sorted([(app_name, names) for (_, app_name, names) in apps])


@presentation.render_for(Admin)
def render(self, h, *args):
    """Display the currently launched application, with their names and their URLs
    """
    with h.div:
        h << h.h2('Active applications')

        with h.ul:
            for (app_name, names) in self.apps:
                with h.li("Application '%s' registered as " % app_name):
                    for (i, name) in enumerate(sorted(names)):
                        if i:
                            h << ' and '
                        h << h.a('/' + name, href='/' + name + ('/' if name else ''))

    return h.root



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --
"""Callbacks manager

Manage the dictionary of the ids / callbacks associations
"""

import random

from nagare.continuation import Continuation


class CallbackLookupError(LookupError):
    pass


def register(model, priority, callback, with_request, render, callbacks):
    """Register a callback

    In:
      - ``model`` -- name of the view which registers this callback (``None`` for the default view)
      - ``priority`` -- type of the callback

        - 0 : <form>.pre_action
        - 1 : action with value (<textarea>, checkbox ...)
        - 2 : action without value (radio button)
        - 3 : <form>.post_action
        - 4 : action with continuation and without value (<a>, submit button ...)
        - 5 : action with continuation and with value (special case for >input type='image'>)

      - ``callback`` -- the action function or method
      - ``with_request`` -- will the request and response objects be passed to the action?
      - ``render`` -- the render function or method

    Out:
      - ``callbacks`` -- dictionary where the keys are the callback ids and the
        values are tuples (callback, with_request, render)

    Return:
      - the callback identifier
    """
    id_ = random.randint(10000000, 99999999)

    # Remember the model, the action and the rendering function
    callbacks[id_] = (model, callback, with_request, render)

    return '_action%d%08d' % (priority, id_)


def clean(old, new):
    """Keep the old callbacks registered by a view only if this view has not registered new callbacks

    In:
      - ``old`` -- the old registered callbacks
      - ``new`` -- the new registered callbacks
    """
    # Retrieve all views which has registered new callbacks
    models = set(callback[0] for callback in new.itervalues())

    # Keep only the old callbacks of a view if no new callbacks were registered
    return dict((k, v) for k, v in old.iteritems() if v[0] not in models)


def process(callbacks, request, response):
    """Call the actions associated to the callback identifiers received

    In:
      - ``callbacks`` -- dictionary where the keys are the callback ids and the
        values are tuples (callback, with_request, render)
      - ``request`` -- the web request object
      - ``response`` -- the web response object

    Return:
      - the render function
    """
    # The structure of a callback identifier is
    # '_action<priority on 1 char><key into the callbacks dictionary>'
    actions = {}

    try:
        for (name, value) in request.params.items():
            if isinstance(value, basestring) and value.startswith('_action'):
                # For the radio buttons, the callback identifier is the value,
                # not the name
                name = value

            if name and name.startswith('_action'):
                v = actions.get(name)
                if v is not None:
                    # Multiple values for the same callback are put into a tuple
                    v = v[3]
                    value = (v if isinstance(v, tuple) else (v,)) + (value,)

                actions[name] = ((int(name[7]), len(actions)), int(name[8:16]), name, value)
    except ValueError:
        raise CallbackLookupError(name[8:])

    render = None

    for ((callback_type, _), name, param, value) in sorted(actions.values()):
        try:
            (model, f, with_request, render) = callbacks[name]
        except KeyError:
            raise CallbackLookupError(name)

        if f is None:
            continue

        # ``callback_type``:
        #
        # 0 : <form>.pre_action
        # 1 : action with value (<textarea>, checkbox ...)
        # 2 : action without value (radio button)
        # 3 : <form>.post_action
        # 4 : action with continuation and without value (<a>, submit button ...)
        # 5 : action with continuation and with value (special case for <input type='image'>)

        if with_request:
            if callback_type == 1:
                f(request, response, value)
            elif callback_type == 4:
                Continuation(f, request, response)
            elif callback_type == 5:
                if param.endswith(('.x', '.y')):
                    Continuation(f, request, response, param.endswith('.y'), int(value))
            else:  # 0, 2, 3
                f(request, response)
        else:
            if callback_type == 1:
                f(value)
            elif callback_type == 4:
                Continuation(f)
            elif callback_type == 5:
                if param.endswith(('.x', '.y')):
                    Continuation(f, param.endswith('.y'), int(value))
            else:  # 0, 2, 3
                f()

    return render



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Helper to validate a configuration"""

import configobj
from validate import Validator


def _validate(filename, config):
    """Validate a ``ConfigObj`` object

    In:
      -  ``filename`` -- the path to the configuration file
      - ``config`` -- the ``ConfigObj`` object, created from the configuration file

    Return:
      - yield the error messages
    """
    errors = configobj.flatten_errors(config, config.validate(Validator(), preserve_errors=True))

    for (sections, name, error) in errors:
        yield 'file "%s", section "[%s]", parameter "%s": %s' % (filename, ' / '.join(sections), name, error)


def validate(filename, config, error):
    """Validate a ``ConfigObj`` object

    In:
      -  ``filename`` -- the path to the configuration file
      - ``config`` -- the ``ConfigObj`` object, created from the configuration file
      - ``error`` -- the function to call in case of configuration errors

    Return:
      - is the configuration valid ?
    """
    errors = list(_validate(filename, config))
    if errors:
        error('\n'.join(errors))
        return False

    return True



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""XSL renderer"""

from __future__ import with_statement

from nagare.namespaces import xml
from nagare.namespaces.xml import TagProp


# ----------------------------------------------------------------------------------

class Renderer(xml.XmlRenderer):
    """ The XSL renderer
    """

    content_type = 'text/xsl'

    # The XSL tags
    # ------------

    apply_imports = TagProp('apply-imports', set())
    apply_templates = TagProp('apply-templates', set(('select', 'mode')))
    attribute = TagProp('attribute', set(('name', 'namespace')))
    attribute_set = TagProp('attribute-set', set(('name', 'use-attribute-sets')))
    call_template = TagProp('call-template', set(('name',)))
    choose = TagProp('choose', set())
    comment = TagProp('comment', set())
    copy = TagProp('copy', set(('use-attribute-sets',)))
    copy_of = TagProp('copy-of', set(('select',)))
    decimal_format = TagProp('decimal-format', set((
        'name', 'decimal-separator', 'grouping-separator', 'infinity',
        'minus-sign', 'NaN', 'percent', 'per-mille', 'zero-digit',
        'pattern-separator'
    )))
    element = TagProp('element', set(('name', 'namespace', 'use-attribute-sets')))
    fallback = TagProp('fallback', set(('select',)))
    if_ = TagProp('if', set(('test',)))
    import_ = TagProp('import', set(('href',)))
    include = TagProp('include', set(('href',)))
    for_each = TagProp('for-each', set(('select',)))
    key = TagProp('key', set(('name', 'match', 'use')))
    message = TagProp('message', set(('terminate',)))
    namespace_alias = TagProp('namespace-alias', set(('stylesheet-prefix',)))
    number = TagProp('number', set(('level', 'count', 'from', 'value', 'format',
                                    'lang', 'letter-value', 'grouping-separator',
                                    'grouping-size')))
    otherwise = TagProp('otherwise', set())
    output = TagProp('output', set(('method', 'version', 'encoding',
                                    'omit-xml-declaration', 'standalone',
                                    'doctype-public', 'doctype-system',
                                    'cdata-section-elements', 'indent',
                                    'media-type')))
    param = TagProp('param', set(('name', 'select')))
    preserve_space = TagProp('preserve-space', set(('elements',)))
    processing_instruction = TagProp('processing-instruction', set(('name',)))
    sort = TagProp('sort', set(('select', 'lang', 'data-type', 'order', 'case-order')))
    strip_space = TagProp('strip-space', set(('elements',)))
    stylesheet = TagProp('stylesheet', set(('version', 'id',
                                            'extension-element-prefixes',
                                            'exclude-result-prefixes')))
    template = TagProp('template', set(('match', 'name', 'priority', 'mode')))
    transform = TagProp('transform', set((
        'version', 'id',
        'extension-element-prefixes',
        'exclude-result-prefixes'
    )))
    text = TagProp('text', set())
    value_of = TagProp('value-of', set(('select', 'disable-output-escaping')))
    variable = TagProp('variable', set(('name', 'select')))
    when = TagProp('when', set(('test',)))
    with_param = TagProp('with-param', set(('name', 'select')))


# ----------------------------------------------------------------------------------

"""
<xsl:transform xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
  <xsl:output method="html" encoding="utf-8"/>
  <xsl:template match="/">
    <xsl:copy-of select="."/>
  </xsl:template>
</xsl:transform>
"""

if __name__ == '__main__':
    from nagare.namespaces import xhtml

    x = Renderer()
    x.namespaces = {'xsl': 'http://www.w3.org/1999/XSL/Transform'}
    x.default_namespace = 'xsl'

    xsl = x.stylesheet(
        x.output(encoding="utf-8"),
        x.template(
            x.copy_of(select="."), match="/"
        )
    )

    print xsl.write_xmlstring(xml_declaration=True, pretty_print=True)

    h = xhtml.Renderer()

    page = h.html(h.h1('Hello'), h.h2('world'))

    print page.write_xmlstring(pretty_print=True)

    r = page.getroottree().xslt(xsl)

    print str(r)
    print



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Variables with a functional interface:

  - ``v()`` -- return the value of ``v``
  - ``v(x)`` -- set the value of ``v`` to ``x``

For example:

  ``v = v + 1`` becomes ``v(v()+1)``

Handy into the lambda expressions
"""

_marker = object()


class Var(object):
    """Functional variables
    """
    def __init__(self, v=None):
        """Initialisation

        In:
          - ``v`` -- initial value
        """
        self.input = v

    def get(self):
        """Return the value

        Return:
          - the value
        """
        return self.input

    def set(self, v):
        """Set the value

        Return:
          - the value
        """
        self.input = v

    def __call__(self, v=_marker):
        """Return or set the value

        In:
          - ``v`` -- if given, ``v`` becomes the new value

        Return:
          - the variable value
        """
        if id(v) != id(_marker):
            self.set(v)

        return self.get()

    def render(self, renderer):
        """When directly put into a XML tree, render its value

        In:
          - ``renderer`` -- the current renderer

        Return:
          - the variable value
        """
        return self.get()

    def __unicode__(self):
        return unicode(self.get())



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""A ``Continuation()`` object captures an execution context

Calling ``switch()`` on a continuation permutes the current execution context
and the captured one, thus resuming where the context was captured.
"""

Tasklet = None

try:
    from _continuation import continulet
except ImportError:
    try:
        import stackless
    except ImportError:
        # CPython
        # -------

        has_continuation = False

        def get_current():
            """Return the current execution context
            """
            return Continuation(lambda: None)

        class Continuation(object):
            """A ``Continuation()`` object launches a function in a new execution context"""

            def __init__(self, f, *args, **kw):
                """Create a new execution context where ``f`` is launched.

                This new execution context became the current one.

                In:
                  - ``f`` -- function to call
                  - ``args``, ``kw`` -- ``f`` arguments
                """
                # CPython: don't create an execution context. Only call the function
                f(*args, **kw)

            def switch(self, value=None):
                """Permute this execution context with the current one

                In:
                  - ``value`` - value returned to the captured execution context
                """
                # No continuation objects in CPython :(
                raise NotImplementedError('Stackless Python or PyPy is needed to create continuations')
    else:
        # Stackless Python
        # ----------------

        has_continuation = True
        Tasklet = stackless.tasklet

        def get_current():
            """Return the current execution context
            """
            return Channel()

        def Continuation(f, *args, **kw):
            """Create a new execution context where ``f`` is launched.

            This new execution context became the current one.

            In:
              - ``f`` -- function to call
              - ``args``, ``kw`` -- ``f`` arguments
            """
            stackless.tasklet(f)(*args, **kw).run()

        class Channel(stackless.channel):
            def switch(self, value=None):
                """Permute this execution context with the current one

                In:
                  - ``value`` - value returned to the captured execution context
                """
                if self.balance:
                    self.send(value)
                else:
                    return self.receive()
else:
    # PyPy
    # ----

    import threading

    has_continuation = True
    _tls = threading.local()

    def get_current():
        """Return the current execution context
        """
        return getattr(_tls, 'current', None)

    def set_current(cont):
        """Set the current execution context
        """
        _tls.current = cont

    class Continuation(continulet):
        """A ``Continuation()`` object launches a function in a new execution context"""

        def __init__(self, f, *args, **kw):
            """Create a new execution context where ``f`` is launched.

            This new execution context became the current one.

            In:
              - ``f`` -- function to call
              - ``args``, ``kw`` -- ``f`` arguments
            """
            super(Continuation, self).__init__(lambda cont: f(*args, **kw))
            self.switch()

        def switch(self, value=None):
            """Permute this execution context with the current one

            In:
              - ``value`` - value returned to the captured execution context
            """
            previous = get_current()
            set_current(self)
            r = super(Continuation, self).switch(value)
            set_current(previous)
            return r


def call_wrapper(action, *args, **kw):
    """A wrapper that creates a continuation and calls a function.

    It's necessary to wrapper a callable that do directly or indirectly a
    ``comp.call(o)`` into such a ``call_wrapper``.

    .. note::
        The actions your registered on the ``<a>`` tags or on the submit buttons
        are already wrapped for you.

    In:
      - ``action`` -- a callable. It will be called, wrapped into a new continuation,
        with the ``args`` and ``kw`` parameters.
      - ``args`` -- positional parameters of the callable
      - ``kw`` -- keywords parameters of the callable
    """
    Continuation(action, *args, **kw)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Generic methods to associate views and URLs to objects"""

import types

from peak.rules import when
from webob.exc import HTTPNotFound


class ModelError(LookupError):
    pass


def render(self, renderer, comp, model):
    """Generic method to associate views to an object

    The views are implementation functions of this generic method

    This default implementation raises an exception

    In:
      - ``self`` -- the object
      - ``renderer`` -- the renderer
      - ``comp`` -- the component
      - ``model`` -- the name of the view
    """
    if model is not None:
        raise ModelError('No model "%s" for %r' % (model, self))
    else:
        raise ModelError('No default model for %r' % self)


def render_for_cond(cond):
    """Decorator helper to register a view for a given condition

    In:
      - ``cond`` -- the condition

    Return:
      - closure
    """
    return when(render, cond)


def render_for(cls, model=None):
    """Decorator helper to register a view for a class of objects

    In:
      - ``cls`` -- the class
      - ``model`` -- the name of the view

    Return:
      - a closure
    """
    if model is not None:
        cond = 'isinstance(self, %s) and (model=="%s")' % (cls.__name__, model)
    else:
        # No name give, dispatch only on the arguments type
        cond = (cls, object, object, types.NoneType)

    return render_for_cond(cond)


@when(render, (object, object, object, int))
def render(self, renderer, comp, model):
    return render(self, renderer, comp, None)


# ---------------------------------------------------------------------------

def init(self, url, comp, http_method, request):
    """Generic method to initialized an object from a URL

    In:
      - ``self`` -- the object
      - ``url`` -- the URL
      - ``comp`` -- the component
      - ``http_method`` -- the HTTP method
      - ``request`` -- the web request object
    """
    raise HTTPNotFound()


def init_for(cls, cond=None):
    """Decorator helper to register an URL for a class of objects

    In:
      - ``cls`` -- the class
      - ``cond`` -- a generic condition

    Return:
      - a closure
    """
    if cond is not None:
        cond = "isinstance(self, %s) and (%s)" % (cls.__name__, cond)
    else:
        # No condition given, dispatch on the class
        cond = (cls,)

    return when(init, cond)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""A ``WSGIApp`` object is an intermediary object between a publisher and the
root component of the application

A ``WSGIApp`` conforms to the WSGI interface and has a component factory. So,
each time the ``WSGIApp`` receives a request without a session id or with an
expired session id, it creates a new root component and a new session.
"""
from __future__ import with_statement

import sys
import os

import webob
from webob import exc, acceptparse

from nagare import component, presentation, serializer, database, top, security, log, comet, i18n, local
from nagare.security import dummy_manager
from nagare.callbacks import CallbackLookupError
from nagare.callbacks import process as process_callbacks
from nagare.namespaces import xhtml

from nagare.sessions import ExpirationError, SessionSecurityError


# ---------------------------------------------------------------------------

class Request(webob.Request):
    @property
    def is_xhr(self):
        return super(Request, self).is_xhr or ('_a' in self.params)

    def create_redirect_response(self, location=None):
        r = webob.exc.HTTPTemporaryRedirect(location=location)
        if self.is_xhr:
            r.status = exc.HTTPServiceUnavailable.code

        return r


class Response(webob.Response):
    def __init__(self, accept):
        super(Response, self).__init__(headerlist=[])

        accept = acceptparse.Accept(accept)
        self.xml_output = accept.best_match(('text/html', 'application/xhtml+xml')) == 'application/xhtml+xml'

        self.content_type = ''
        self.doctype = None


# ---------------------------------------------------------------------------

class WSGIApp(object):
    request_factory = Request
    response_factory = Response
    renderer_factory = xhtml.Renderer   # Default renderer

    def __init__(self, root_factory):
        """Initialization

        In:
          - ``root_factory`` -- function to create the application root component
        """
        self.root_factory = root_factory

        self.static_path = ''
        self.static_url = ''
        self.data_path = ''
        self.databases = []
        self.name = ''
        self.project_name = ''
        self.redirect_after_post = False
        self.always_html = True
        self.sessions = None
        self.last_exception = None

        self.security = dummy_manager.Manager()

        self.set_default_locale(i18n.Locale())

    def set_config(self, config_filename, config, error):
        """Read the configuration parameters

        In:
          - ``config_filename`` -- the path to the configuration file
          - ``config`` -- the ``ConfigObj`` object, created from the configuration file
          - ``error`` -- the function to call in case of configuration errors
        """
        self.name = config['application']['name']
        self.redirect_after_post = config['application']['redirect_after_post']
        self.always_html = config['application']['always_html']

    def set_static_path(self, static_path):
        """Register the directory of the static contents

        In:
          - ``static_path`` -- the directory where the static contents of the application
            are located
        """
        self.static_path = static_path

    def set_static_url(self, static_url):
        """Register the url of the static contents

        In:
          - ``static_url`` -- the url of the static contents of the application
        """
        self.static_url = static_url

    def set_data_path(self, data_path):
        """Register the directory of the data

        In:
          - ``data_path`` -- the directory where the data of the application are located
        """
        self.data_path = data_path

    def set_publisher(self, publisher):
        """Register the publisher

        In:
          - ``publisher`` -- the publisher of the application
        """
        pass

    def set_sessions_manager(self, sessions_manager):
        """Register the sessions manager

        In:
          - ``sessions_manager`` -- the sessions manager
        """
        self.sessions = sessions_manager

    def set_databases(self, databases):
        """Register the databases properties

        In:
          - ``databases`` -- the SQLAlchemy metadata objects and the database engines settings
        """
        self.databases = databases

    def set_project(self, name):
        """The application distribution name

        In:
          - ``project_name`` -- name of the distutils distribution where the app is located
        """
        self.project_name = name

    def set_default_locale(self, locale):
        """Register the default locale

        In:
          - ``locale`` -- the default locale
        """
        self.default_locale = locale

    def set_locale(self, locale):
        """Set the locale of the request scope

        In:
          - ``locale`` -- the locale
        """
        if not locale.has_translation_directory():
            locale.add_translation_directory(os.path.join(self.data_path, 'locale'))

        i18n.set_locale(locale)

    def start(self):
        """Call after each process start
        """
        for database_settings in self.databases:
            database.set_metadata(*database_settings)

    # -----------------------------------------------------------------------

    def on_bad_http_method(self, request, response):
        """A HTTP request other than a GET or PUT was received

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object

        Return:
          - raise a ``webob.exc`` object, used to generate the response to the browser
        """
        raise exc.HTTPMethodNotAllowed()

    def on_incomplete_url(self, request, response):
        """An URL without an application name was received

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object

        Return:
          - raise a ``webob.exc`` object, used to generate the response to the browser
        """
        raise exc.HTTPMovedPermanently(add_slash=True)

    def on_expired_session(self, request, response):
        """The session or state id received is expired

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object

        Return:
          - raise a ``webob.exc`` object, used to generate the response to the browser
        """
        raise exc.HTTPMovedPermanently()
    on_session_expired = on_expired_session

    def on_invalid_session(self, request, response):
        """The secure id received is invalid

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object

        Return:
          - raise a ``webob.exc`` object, used to generate the response to the browser
        """
        raise request.create_redirect_response()

    def on_back(self, request, response, h, output):
        """The user used the back button

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object
          - ``h`` -- the current renderer
          - ``output`` -- the tree for the refreshed page

        Return:
          - a tree
        """
        return output

    def on_callback_lookuperror(self, request, response, async):
        """A callback was not found

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object
          - ``async`` -- is an XHR request ?
        """
        if not async:
            raise

        # As the XHR requests use the same continuation, a callback
        # can be not found (i.e deleted by a previous XHR)
        # In this case, do nothing
        return lambda h: ''

    def on_after_post(self, request, response, ids):
        """Generate a redirection after a POST

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object
          - ``ids`` -- identifiers to put into the generated redirection URL

        Return:
          - a ``webob.exc`` object, used to generate the response to the browser
        """
        return exc.HTTPSeeOther(location=request.path_url + '?' + '&'.join(ids))

    def on_exception(self, request, response):
        """Method called when an exception occurs

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object

        Return:
          - a ``webob`` response object
        """
        raise

    # -----------------------------------------------------------------------

    def create_request(self, environ):
        """Create the ``webob.Request`` object

        In:
            - ``environ`` -- dictionary of the received elements from the browser

        Return:
            - a ``webob.Request`` object
        """
        return self.request_factory(environ, charset='utf-8')

    def create_response(self, request, accept):
        """Create the ``webob.Response`` object

        In:
            - ``request`` -- the ``webob.Request`` object
            - ``accept`` -- the ``Accept`` HTTP header to use

        Return:
            - a ``webob.Response`` object
        """
        return self.response_factory(accept)

    def create_root(self, *args, **kw):
        """Create the application root component

        Return:
          - the root component
        """
        return self.root_factory(*args, **kw)

    def create_renderer(self, async, session, request, response):
        """Create the initial renderer (the root of all the used renderers)

        In:
          - ``async`` -- is an XHR request ?
          - ``session`` -- the session
          - ``request`` -- the web request object
          - ``response`` -- the web response object
        """
        renderer = self.renderer_factory(
            None,
            session,
            request, response,
            self.static_url, self.static_path,
            request.script_name
        )

        if async:
            renderer = renderer.AsyncRenderer(
                None,
                session,
                request, response,
                self.static_url, self.static_path,
                request.script_name,
                async_header=True
            )

        return renderer

    def start_request(self, root, request, response):
        """A new request is received, setup its dedicated environment

        In:
          - ``root`` -- the application root component
          - ``request`` -- the web request object
          - ``response`` -- the web response object
        """
        security.set_manager(self.security)  # Set the security manager
        security.set_user(self.security.create_user(request, response))  # Create the User object

        self.set_locale(self.default_locale)  # Set the default Locale

    # Processing phase
    def _phase1(self, root, request, response, callbacks):
        """Phase 1 of the request processing:

          - The callbacks are processed
          - The objects graph can be modified

        In:
          - ``root`` -- the application root component
          - ``request`` -- the web request object
          - ``response`` -- the web response object
          - ``callbacks`` -- the registered callbacks

        Return:
          - function to render the objects graph or ``None``
        """
        return process_callbacks(callbacks or {}, request, response)

    # Rendering phase
    def _phase2(self, output, content_type, doctype, is_xhr, response):
        """Final step of the phase 2

        Phase 2 of the request processing:

          - The object graph is rendered
          - No modification of the objects is allowed

        In:
          - ``output`` -- the rendered content
          - ``content_type`` -- the rendered content type
          - ``doctype`` -- the (optional) doctype
          - ``is_xhr`` -- is the request a XHR request ?

        Out:
          - ``response`` -- the response object
        """
        (response.content_type, response.body) = serializer.serialize(output, content_type, doctype, not is_xhr)
        response.charset = 'utf-8'

    def __call__(self, environ, start_response):
        """WSGI interface

        In:
          - ``environ`` -- dictionary of the received elements
          - ``start_response`` -- callback to send the headers to the browser

        Return:
          - the content to send back to the browser
        """
        local.request.clear()

        # Create the ``WebOb`` request and response objects
        # -------------------------------------------------

        request = self.create_request(environ)
        try:
            request.params, request.url
        except UnicodeDecodeError:
            return exc.HTTPClientError()(environ, start_response)

        response = self.create_response(request, 'text/html' if self.always_html else str(request.accept))

        channel_id = request.params.get('_channel')
        nb = request.params.get('_nb')
        if channel_id and nb:
            if environ['wsgi.multiprocess']:
                response.status = 501  # "Not Implemented"
            else:
                comet.channels.connect(channel_id, int(nb), environ['wsgi.input'].file.fileno(), response)

            return response(environ, start_response)

        xhr_request = request.is_xhr

        state = None
        self.last_exception = None

        log.set_logger('nagare.application.' + self.name)  # Set the dedicated application logger

        # Create a database transaction for each request
        with database.session.begin():
            try:
                # Phase 1
                # -------

                # Test the request validity
                if not request.path_info:
                    self.on_incomplete_url(request, response)

                try:
                    state = self.sessions.get_state(request, response, xhr_request)
                except ExpirationError:
                    self.on_expired_session(request, response)
                except SessionSecurityError:
                    self.on_invalid_session(request, response)

                state.acquire()

                try:
                    root, callbacks = state.get_root() or (self.create_root(), None)
                except ExpirationError:
                    self.on_expired_session(request, response)
                except SessionSecurityError:
                    self.on_invalid_session(request, response)

                self.start_request(root, request, response)

                if callbacks is None:
                    # New state
                    request.method = request.params.get('_method', request.method)
                    if request.method not in ('GET', 'POST'):
                        self.on_bad_http_method(request, response)

                    url = request.path_info.strip('/')
                    if url:
                        # If a URL is given, initialize the objects graph with it
                        presentation.init(root, tuple(url.split('/')), None, request.method, request)

                try:
                    render = self._phase1(root, request, response, callbacks)
                except CallbackLookupError:
                    render = self.on_callback_lookuperror(request, response, xhr_request)
            except exc.HTTPException, response:
                # When a ``webob.exc`` object is raised during phase 1, skip the
                # phase 2 and use it as the response object
                pass
            except Exception:
                self.last_exception = (request, sys.exc_info())
                response = self.on_exception(request, response)
            else:
                # Phase 2
                # -------

                # If the ``redirect_after_post`` parameter of the ``[application]``
                # section is `True`` (the default), conform to the PRG__ pattern
                #
                # __ http://en.wikipedia.org/wiki/Post/Redirect/GetPRG

                try:
                    if (request.method == 'POST') and not xhr_request and self.redirect_after_post:
                        use_same_state = True
                        response = self.on_after_post(request, response, state.sessionid_in_url(request, response))
                    else:
                        use_same_state = xhr_request

                        # Create a new renderer
                        renderer = self.create_renderer(xhr_request, state, request, response)
                        # If the phase 1 has returned a render function, use it
                        # else, start the rendering by the application root component
                        output = render(renderer) if render else root.render(renderer)

                        if state.back_used:
                            output = self.on_back(request, response, renderer, output)

                        if not xhr_request:
                            output = top.wrap(renderer.content_type, renderer, output)

                        self._phase2(output, renderer.content_type, renderer.doctype, xhr_request, response)

                    # Store the state
                    state.set_root(use_same_state, root)

                    security.get_manager().end_rendering(request, response, state)
                except exc.HTTPException, response:
                    # When a ``webob.exc`` object is raised during phase 2, stop immediately
                    # use it as the response object
                    pass
                except Exception:
                    self.last_exception = (request, sys.exc_info())
                    response = self.on_exception(request, response)
            finally:
                if state:
                    state.release()

        return response(environ, start_response)


# ---------------------------------------------------------------------------

def create_WSGIApp(app, with_component=True):
    """Helper function to create a WSGIApp

    If ``app`` is not a ``WSGIApp``, it's wrap into a ``WSGIApp``. And, if
    ``with_component`` is True, each time a new root object is created, it's
    automatically wrap into a ``component.Component``.

    In:
      - ``app`` -- the application root component factory
      - ``with_component`` -- wrap a new root object into a component
    """
    if not isinstance(app, WSGIApp):
        if with_component:
            def wrap_in_component(factory):
                o = factory()
                if not isinstance(o, component.Component):
                    o = component.Component(o)
                return o

            app = lambda app=app: wrap_in_component(app)  # noqa: E731

        app = WSGIApp(app)

    return app



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

from __future__ import with_statement

import csv
import os
from types import ListType

from lxml import etree as ET

from nagare.namespaces import xml


def append_text_test1():
    """ XML namespace unit test - append_text Append text to an empty node

    In:
      - <node/>

    Out:
      - <node>test</node>
    """
    x = xml.Renderer()

    node = x.node()
    node.append_text('test')
    assert node.text == 'test'
    assert node.write_xmlstring() == '<node>test</node>'


def append_text_test2():
    """ XML namespace unit test - append_text - Append text to node with text child

    In:
      - <node>test1</node>

    Out:
      - <node>test1test2</node>
    """
    x = xml.Renderer()

    node = x.node('test1')
    node.append_text('test2')
    assert node.text == 'test1test2'
    assert node.write_xmlstring() == '<node>test1test2</node>'


def append_text_test3():
    """ XML namespace unit test - append_text - Append text to node with node child

    In:
      - <node><child/></node>

    Out:
      - <node><child/>test</node>
    """
    x = xml.Renderer()

    node = x.node(x.child())
    node.append_text('test')
    assert node.getchildren()[0].tail == 'test'
    assert node.write_xmlstring() == '<node><child/>test</node>'


def append_text_test4():
    """ XML namespace unit test - append_text - Append text to node with text & node children

    In:
      - <node>test1<child/></node>

    Out:
      - <node>test1<child/>test2</node>
    """
    x = xml.Renderer()

    node = x.node(['test1', x.child()])
    node.append_text('test2')
    assert node.text == 'test1'
    assert node.getchildren()[0].tail == 'test2'
    assert node.write_xmlstring() == '<node>test1<child/>test2</node>'


def add_child_test1():
    """ XML namespace unit test - add_child - add text

    In:
      - <node/>

    Out:
      - <node>test</node>
    """
    x = xml.Renderer()

    node = x.node('test')
    assert node.text == 'test'
    assert node.write_xmlstring() == '<node>test</node>'


def add_child_test2():
    """ XML namespace unit test - add_child - add node

    In:
      - <node/>

    Out:
      - <node><child/></node>
    """
    x = xml.Renderer()

    child = x.child()
    node = x.node(child)
    assert node.getchildren()[0] == child
    assert node.write_xmlstring() == '<node><child/></node>'


def add_child_test3():
    """ XML namespace unit test - add_child - add dictionnary

    In:
      - <node/>

    Out:
      - <node test="test"/>
    """
    x = xml.Renderer()

    node = x.node({'test': 'test'})
    assert node.attrib == {'test': 'test'}
    assert node.write_xmlstring() == '<node test="test"/>'


def add_child_test4():
    """ XML namespace unit test - add_child - add attribute with keyword

    In:
      - <node/>

    Out:
      - <node test="test"/>
    """
    x = xml.Renderer()

    node = x.node(test='test')
    assert node.attrib == {'test': 'test'}
    assert node.write_xmlstring() == '<node test="test"/>'


def add_child_test5():
    """ XML namespace unit test - add_child - add tuple (text, node, text, dict)

    In:
      - <node/>

    Out:
      - <node test="test">test<child/>test</node>
    """
    x = xml.Renderer()

    child = x.child()
    node = x.node(('test', child, 'test', {'test': 'test'}))

    assert node.attrib == {'test': 'test'}
    assert node.text == 'test'
    assert node.getchildren()[0] == child
    assert node.getchildren()[0].tail == 'test'
    assert node.write_xmlstring() == '<node test="test">test<child/>test</node>'


def add_child_test6():
    """ XML namespace unit test - add_child - add list [text, node, text, dict]

    In:
      - <node/>

    Out:
      - <node test="test">test<child/>test</node>
    """
    x = xml.Renderer()

    child = x.child()
    node = x.node(['test', child, 'test', {'test': 'test'}])

    assert node.attrib == {'test': 'test'}
    assert node.text == 'test'
    assert node.getchildren()[0] == child
    assert node.getchildren()[0].tail == 'test'
    assert node.write_xmlstring() == '<node test="test">test<child/>test</node>'


def add_child_test7():
    """ XML namespace unit test - add_child - add int

    In:
      - <node/>

    Out:
      - <node>42</node>
    """
    x = xml.Renderer()

    node = x.node(42)
    assert node.text == '42'
    assert node.write_xmlstring() == '<node>42</node>'


def add_child_test8():
    """ XML namespace unit test - add_child - add float

    In:
      - <node/>

    Out:
      - <node>0.1</node>
    """
    x = xml.Renderer()

    node = x.node(0.1)
    assert node.text == '0.1'
    assert node.write_xmlstring() == '<node>0.1</node>'


def add_child_test9():
    """ XML namespace unit test - add_child - add dictionnary with python keyword

    In:
      - <node/>

    Out:
      - <node class="test"/>
    """
    x = xml.Renderer()

    node = x.node({'class_': 'test'})
    assert node.attrib == {'class': 'test'}
    assert node.write_xmlstring() == '<node class="test"/>'


def add_child_test10():
    """ XML namespace unit test - add_child - add attribute with python keyword

    In:
      - <node/>

    Out:
      - <node class="test"/>
    """
    x = xml.Renderer()

    node = x.node(class_='test')
    assert node.attrib == {'class': 'test'}
    assert node.write_xmlstring() == '<node class="test"/>'


def add_child_test11():
    """ XML namespace unit test - add_child - add object instance and raise exception

    In:
      - <node/>

    Out:
      - Exception
    """
    x = xml.Renderer()

    try:
        x.node(object())
    except Exception:
        assert True
    else:
        assert False


def replace_test1():
    """ XML namespace unit test - replace - replace simple node by node

    In:
      - <node><child1/></node>

    Out:
      - <node><child2/></node>
    """
    x = xml.Renderer()

    child1 = x.child1()
    node = x.node(child1)
    assert node.getchildren()[0] == child1
    child2 = x.child2()
    child1.replace(child2)
    assert node.getchildren()[0] == child2
    assert node.write_xmlstring() == '<node><child2/></node>'


def replace_test2():
    """ XML namespace unit test - replace - replace simple node with text before by node

    In:
      - <node><child1/></node>

    Out:
      - <node><child2/></node>
    """
    x = xml.Renderer()

    child1 = x.child1()
    node = x.node('test', child1)
    assert node.getchildren()[0] == child1
    child2 = x.child2()
    child1.replace(child2)
    assert node.getchildren()[0] == child2
    assert node.write_xmlstring() == '<node>test<child2/></node>'


def replace_test3():
    """ XML namespace unit test - replace - replace simple node with text after by node

    In:
      - <node><child1/></node>

    Out:
      - <node><child2/></node>
    """
    x = xml.Renderer()

    child1 = x.child1()
    node = x.node(child1, 'test')
    assert node.getchildren()[0] == child1
    child2 = x.child2()
    child1.replace(child2)
    assert node.getchildren()[0] == child2
    assert node.write_xmlstring() == '<node><child2/>test</node>'


def replace_test4():
    """ XML namespace unit test - replace - replace simple node by text

    In:
      - <node><child1/></node>

    Out:
      - <node>test</node>
    """
    x = xml.Renderer()

    child1 = x.child1()
    node = x.node(child1)
    assert node.getchildren()[0] == child1
    child1.replace('test')
    assert node.text == 'test'
    assert node.write_xmlstring() == '<node>test</node>'


def replace_test5():
    """ XML namespace unit test - replace - replace simple node by node + text

    In:
      - <node><child1/></node>

    Out:
      - <node>test<child2/></node>
    """
    x = xml.Renderer()

    child1 = x.child1()
    node = x.node(child1)
    assert node.getchildren()[0] == child1
    child2 = x.child2()
    child1.replace('test', child2)
    assert node.text == 'test'
    assert node.write_xmlstring() == '<node>test<child2/></node>'


def replace_test6():
    """ XML namespace unit test - replace - replace root node

    In:
      - <node/>

    Out:
      - <node/>
    """
    x = xml.Renderer()

    node = x.node()
    node.replace('test')
    assert node.write_xmlstring() == '<node/>'


xml_test1_in = """
    <node xmlns:meld="http://www.plope.com/software/meld3">
    <child meld:id="child"/>
    </node>
"""


def repeat_test1():
    """ XML namespace unit test - repeat - Repeat with 2 simple text, use childname argument"""
    x = xml.Renderer()
    node = x.parse_xmlstring(xml_test1_in)
    iterator = node.repeat(['test1', 'test2'], childname='child')

    for child, value in iterator:
        child.append_text(value)
    assert [child.text for child in node.getchildren()] == ['test1', 'test2']


def repeat_test2():
    """ XML namespace unit test - repeat - Repeat with 2 simple text, don't use childname argument"""
    x = xml.Renderer()
    node = x.parse_xmlstring(xml_test1_in)
    child = node.findmeld('child')
    iterator = child.repeat(['test1', 'test2'])

    for child, value in iterator:
        child.append_text(value)
    assert [child.text for child in node.getchildren()] == ['test1', 'test2']


def repeat_test3():
    """ XML namespace unit test - repeat - findmeld in repeat loop """
    h = xml.Renderer()

    xhtml_tree_2 = '<div xmlns:meld="http://www.plope.com/software/meld3"><ul><li meld:id="entry"><span meld:id="count">count</span></li></ul></div>'

    root = h.parse_xmlstring(xhtml_tree_2, fragment=True)[0]

    for (elem, count) in root.repeat(range(2), 'entry'):
        elem.findmeld('count').fill(count)

    h << root

    assert h.root.write_xmlstring() == '<div xmlns:meld="http://www.plope.com/software/meld3"><ul><li meld:id="entry"><span meld:id="count">0</span></li><li meld:id="entry"><span meld:id="count">1</span></li></ul></div>'


def root_test1():
    """ XML namespace unit test - root - one element """
    x = xml.Renderer()
    x << x.node()

    assert not isinstance(x.root, ListType)


def root_test2():
    """ XML namespace unit test - root - two elements """
    x = xml.Renderer()
    x << x.node()
    x << x.node()

    assert isinstance(x.root, ListType)


def parse_xmlstring_test1():
    """ XML namespace unit test - parse_xmlstring - good encoding """
    try:
        x = xml.Renderer()
        f = open(os.path.join(os.path.dirname(__file__), 'test_xmlns_1.xml'))
        x.parse_xmlstring(f.read())
        f.close()
    except UnicodeDecodeError:
        assert False
    else:
        assert True


def parse_xmlstring_test2():
    """ XML namespace unit test - parse_xmlstring - bad encoding """
    try:
        x = xml.Renderer()
        f = open(os.path.join(os.path.dirname(__file__), 'iso-8859.xml'))
        x.parse_xml(f, encoding='utf-8')
        f.close()
    except ET.XMLSyntaxError:
        assert True
    else:
        assert False


xml_fragments_1 = """leading_text<fragment1></fragment1>text<fragment2></fragment2>"""


def parse_xmlstring_test3():
    """ XML namespace unit test - parse_xmlstring - parse fragment xml with fragment flag """
    x = xml.Renderer()
    roots = x.parse_xmlstring(xml_fragments_1, fragment=True)
    assert roots[0] == 'leading_text'
    assert roots[1].write_xmlstring() == "<fragment1/>text"
    assert roots[1].tail == "text"
    assert roots[2].write_xmlstring() == "<fragment2/>"


def parse_xmlstring_test4():
    """ XML namespace unit test - parse_xmlstring - parse xml tree with fragment flag """
    x = xml.Renderer()
    roots = x.parse_xmlstring(xml_fragments_1, fragment=True, no_leading_text=True)
    assert roots[0].write_xmlstring() == "<fragment1/>text"
    assert roots[0].tail == "text"
    assert roots[1].write_xmlstring() == "<fragment2/>"


xml_tree_1 = "<a>text</a>"


def parse_xmlstring_test5():
    """ XML namespace unit test - parse_xmlstring - Test parse child type """
    x = xml.Renderer()
    root = x.parse_xmlstring(xml_tree_1)
    assert type(root) == xml._Tag


def parse_xml_test1():
    """ XML namespace unit test - parse_xmlstring - good encoding """
    try:
        x = xml.Renderer()
        x.parse_xml(os.path.join(os.path.dirname(__file__), 'test_xmlns_1.xml'))
    except UnicodeDecodeError:
        assert False
    else:
        assert True


def findmeld_test1():
    """ XML namespace unit test - find_meld - one element """
    x = xml.Renderer()
    node = x.parse_xmlstring(xml_test1_in)
    child = node.findmeld('child')

    assert child is not None


def findmeld_test2():
    """ XML namespace unit test - find_meld - zero element """
    x = xml.Renderer()
    node = x.parse_xmlstring("""<node xmlns:meld="http://www.plope.com/software/meld3"></node>""")
    child = node.findmeld('child')
    assert child is None


def findmeld_test3():
    """ XML namespace unit test - find_meld - zero element and default argument """
    x = xml.Renderer()
    node = x.parse_xmlstring("""<node xmlns:meld="http://www.plope.com/software/meld3"></node>""")
    child = node.findmeld('child', 'test')
    assert child == 'test'


# Test for XML namespace

xml_test2_in = """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
    <html xmlns="http://www.w3.org/1999/xhtml"
          xmlns:meld="http://www.plope.com/software/meld3"
          xmlns:bar="http://foo/bar">
      <head>
        <meta content="text/html; charset=ISO-8859-1" http-equiv="content-type" />
        <title meld:id="title">This is the title</title>
      </head>
      <body>
        <div/> <!-- empty tag -->
        <div meld:id="content_well">
          <form meld:id="form1" action="." method="POST">
          <table border="0" meld:id="table1">
            <tbody meld:id="tbody">
              <tr>
                <th>Name</th>
                <th>Description</th>
              </tr>
              <tr meld:id="tr" class="foo">
                <td meld:id="td1">Name</td>
                <td meld:id="td2">Description</td>
              </tr>
            </tbody>
          </table>
          <input type="submit" name="next" value=" Next "/>
          </form>
        </div>
      </body>
    </html>"""


def global_test1():
    """ XML namespace unit test - create xml by procedural way """
    x = xml.Renderer()

    filePath = os.path.join(os.path.dirname(__file__), 'helloworld.csv')

    reader = csv.reader(open(filePath, 'r'))

    with x.helloWorlds:
        for row in reader:
            with x.helloWorld(language=row[0].decode('utf-8')):
                x << row[1].decode('utf-8')

    xmlToTest = x.root.write_xmlstring(xml_declaration=True, pretty_print=True).strip()

    f = open(os.path.join(os.path.dirname(__file__), 'test_xmlns_1.xml'))
    xmlToCompare = f.read()
    f.close()
    assert xmlToTest == xmlToCompare


def global_test2():
    """ XML namespace unit test - create xml by functionnal way """
    x = xml.Renderer()

    filePath = os.path.join(os.path.dirname(__file__), 'helloworld.csv')

    reader = csv.reader(open(filePath, 'r'))

    root = x.helloWorlds(
        [x.helloWorld(row[1].decode('utf-8'), {'language': row[0].decode('utf-8')}) for row in reader]
    )

    xmlToTest = root.write_xmlstring(xml_declaration=True, pretty_print=True).strip()

    f = open(os.path.join(os.path.dirname(__file__), 'test_xmlns_1.xml'))
    xmlToCompare = f.read()
    f.close()
    assert xmlToTest == xmlToCompare


def global_test3():
    """ XML namespace unit test - test parse_xmlstring method """
    x = xml.Renderer()

    f = open(os.path.join(os.path.dirname(__file__), 'test_xmlns_1.xml'))
    root = x.parse_xmlstring(f.read())
    f.close()

    xmlToTest = root.write_xmlstring(xml_declaration=True, pretty_print=True).strip()

    f = open(os.path.join(os.path.dirname(__file__), 'test_xmlns_1.xml'))
    xmlToCompare = f.read()
    f.close()
    assert xmlToTest == xmlToCompare


def global_test4():
    """ XML namespace unit test - meld3 - test findmeld with children affectation """

    x = xml.Renderer()
    root = x.parse_xmlstring(xml_test2_in)

    root.findmeld('title').text = 'My document'
    root.findmeld('form1').set('action', './handler')
    data = (
        {'name': 'Girls', 'description': 'Pretty'},
        {'name': 'Boys', 'description': 'Ugly'},
    )

    iterator = root.findmeld('tr').repeat(data)
    for element, item in iterator:
        td1 = element.findmeld('td1')
        td1.text = item['name']
        element.findmeld('td2').text = item['description']

    assert [elt.text for elt in root.xpath('.//x:td', namespaces={'x': 'http://www.w3.org/1999/xhtml'})] == ['Girls', 'Pretty', 'Boys', 'Ugly']
    assert root[0][1].text == 'My document'
    assert root.xpath('.//x:form', namespaces={'x': 'http://www.w3.org/1999/xhtml'})[0].attrib['action'] == './handler'


def global_test5():
    """ XML namespace unit test - meld3 - test findmeld & replace method """
    x = xml.Renderer()
    root = x.parse_xmlstring(xml_test2_in)

    root.findmeld('title').text = 'My document'
    root.findmeld('form1').set('action', './handler')

    data = (
        {'name': 'Girls', 'description': 'Pretty'},
        {'name': 'Boys', 'description': 'Ugly'},
    )

    children = []
    for elt in data:
        children.append(x.tr([x.td(elt['name']), x.td(elt['description'])], {'class': 'bar'}))

    root.findmeld('tr').replace(children)

    assert root.findall('.//tr')[1].attrib['class'] == 'bar'
    assert [elt.text for elt in root.xpath('.//td')] == ['Girls', 'Pretty', 'Boys', 'Ugly']


def global_test6():
    """ XML namespace unit test - meld3 - test option write_xmlstring method with option pipeline=False """
    x = xml.Renderer()
    root = x.parse_xmlstring(xml_test2_in)

    root.findmeld('title').text = 'My document'
    root.findmeld('form1').set('action', './handler')

    data = (
        {'name': 'Girls', 'description': 'Pretty'},
        {'name': 'Boys', 'description': 'Ugly'},
    )

    children = []
    for elt in data:
        children.append(x.tr([x.td(elt['name']), x.td(elt['description'])], {'class': 'bar'}))

    root.findmeld('tr').replace(children)

    root.write_xmlstring(xml_declaration=True, pretty_print=True, pipeline=False).strip()

    assert root.findmeld('tr') is None
    assert root.findmeld('content_well') is None


def global_test7():
    """ XML namespace unit test - meld3 - test option write_xmlstring method with option pipeline=True """
    x = xml.Renderer()
    root = x.parse_xmlstring(xml_test2_in)

    root.findmeld('title').text = 'My document'
    root.findmeld('form1').set('action', './handler')

    data = (
        {'name': 'Girls', 'description': 'Pretty'},
        {'name': 'Boys', 'description': 'Ugly'},
    )

    children = []
    for elt in data:
        children.append(x.tr([x.td(elt['name']), x.td(elt['description'])], {'class': 'bar'}))

    root.findmeld('tr').replace(children)

    root.write_xmlstring(xml_declaration=True, pretty_print=True, pipeline=False).strip()

    assert root.findmeld('tr') is None
    assert root.findmeld('content_well') is None


def global_test8():
    """ XML namespace unit test - create xml """
    x = xml.Renderer()
    x.namespaces = {'meld': 'http://www.plope.com/software/meld3'}
    data = (
        {'name': 'Girls', 'description': 'Pretty'},
        {'name': 'Boys', 'description': 'Ugly'},
    )

    with x.html:
        with x.head:
            with x.meta:
                x << {'content': 'text/html; charset=ISO-8859-1', 'http-equiv': 'content-type'}
            with x.title.meld_id('title'):
                x << 'My document'
        with x.body:
            with x.div:
                pass
            x << x.comment(' empty tag ')
            with x.div:
                with x.form.meld_id('form1'):
                    x << {'action': './handler'}
                    with x.table:
                        with x.tbody:
                            with x.tr:
                                x << x.th('Name') << x.th('Description')
                            for elt in data:
                                with x.tr.meld_id('tr'):
                                    x << x.td(elt['name']).meld_id('td') << x.td(elt['description']).meld_id('td')
                    with x.input:
                        x << {'type': 'submit', 'name': 'next', 'value': ' Next '}

    assert [elt.text for elt in x.root.xpath('.//td')] == ['Girls', 'Pretty', 'Boys', 'Ugly']
    assert x.root[0][1].text == 'My document'
    assert x.root.xpath('.//form')[0].attrib['action'] == './handler'



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Edge Server-side Include renderer"""

from lxml import etree as ET

from nagare.namespaces import xml
from nagare.namespaces.xml import TagProp

# Official ESI namespace
NS = 'http://www.edge-delivery.org/esi/1.0'


class Renderer(xml.XmlRenderer):
    """ The ESI renderer
    """

    # The ESI tags
    # ------------

    include = TagProp('include', set(('src', 'alt', 'onerror')))
    inline = TagProp('inline', set(('name', 'fetchable')))
    choose = TagProp('choose', set())
    when = TagProp('when', set(('test',)))
    otherwise = TagProp('otherwise', set())
    try_ = TagProp('try', set())
    attempt = TagProp('attempt', set())
    except_ = TagProp('except', set())
    comment = TagProp('comment', set(('text',)))
    remove = TagProp('remove', set())
    vars = TagProp('vars', set())

    def esi(self, text):
        """Generate a ``esi`` comment element

        In:
          - ``text`` -- comment text

        Return:
          - the comment element
        """
        return ET.Comment('esi ' + text)


if __name__ == '__main__':
    from nagare.namespaces import xhtml

    h = xhtml.Renderer()
    s = Renderer()
    s.namespaces = {'esi': NS}
    s.default_namespace = 'esi'

    html = h.html(h.body(h.h1('hello'), s.include(dict(src='http://www.net-ng.com'), s.include(src='http://www.net-ng.com')), h.p('world')))

    print html.write_xmlstring(pretty_print=True)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

# -----------------------------------------------------------------------------

# If the framework haven't the SQLAlchemy or Elixir packages installed,
# the default database session is a dummy one


class DummySession(object):
    def begin(self):
        return self

    __enter__ = __exit__ = clear = query = remove = close = configure = lambda *args, **kw: None


session = DummySession()


def setup_all():
    pass


def add_pickle_hooks(mapper, cls):
    # Dynamically add a ``__getstate__()`` and ``__setstate__()`` method
    # to the SQLAlchemy entities
    if not hasattr(cls, '__getstate__'):
        cls.__getstate__ = entity_getstate

    if not hasattr(cls, '__setstate__'):
        cls.__setstate__ = entity_setstate


try:
    # SQLAlchemy >= 0.7.6
    import sqlalchemy
    from sqlalchemy import event, orm

    event.listen(orm.Mapper, 'mapper_configured', add_pickle_hooks)
except ImportError:
    pass


try:
    # SQLAlchemy <= 0.8.7
    from sqlalchemy import orm

    class Mapper(orm.Mapper):
        def __init__(self, cls, *args, **kw):
            super(Mapper, self).__init__(cls, *args, **kw)
            add_pickle_hooks(self, cls)

    # Hot-patching the SQLAlchemy ``Mapper`` class
    orm.Mapper = Mapper
except ImportError:
    pass


try:
    from sqlalchemy import orm

    # ``sqlalchemy.orm.ScopedSession`` is needed by ``elixir``
    orm.__dict__.setdefault('ScopedSession', orm.scoped_session)

    session = orm.scoped_session(orm.sessionmaker())
except ImportError:
    pass


try:
    from elixir import setup_all, session  # noqa: F811
except ImportError:
    pass


session.configure(autoflush=True, expire_on_commit=True, autocommit=True)
query = session.query


# -----------------------------------------------------------------------------

def entity_getstate(entity):
    """Return the state of an SQLAlchemy entity

    In:
      - ``entity`` -- the SQLAlchemy entity

    Return:
      - the state dictionary
    """
    state = entity._sa_instance_state  # SQLAlchemy managed state

    if state.key:
        attrs = set(state.manager.local_attrs)  # SQLAlchemy managed attributes
        attrs.add('_sa_instance_state')
    else:
        attrs = ()

    # ``d`` is the dictionary of the _not_ SQLAlchemy managed attributes
    d = dict([(k, v) for (k, v) in entity.__dict__.items() if k not in attrs])

    # Keep only the primary key from the SQLAlchemy state
    d['_sa_key'] = state.key[1] if state.key else None

    return d


def entity_setstate(entity, d):
    """Set the state of an SQLAlchemy entity

    In:
      - ``entity`` -- the newly created and not yet initialized SQLAlchemy entity
      - ``d`` -- the state dictionary (created by ``entity_getstate()``)
    """
    # Copy the _not_ SQLAlchemy managed attributes to our entity
    key = d.pop('_sa_key', None)
    entity.__dict__.update(d)

    if key is not None:
        # Fetch a new and initialized SQLAlchemy from the database
        x = session.query(entity.__class__).get(key)
        session.expunge(x)

        # Copy its state to our entity
        entity.__dict__.update(x.__dict__)

        # Adjust the entity SQLAlchemy state
        state = x._sa_instance_state.__getstate__()
        state['instance'] = entity
        entity._sa_instance_state.__setstate__(state)

        # Add the entity to the current database session
        session.add(entity)


# -----------------------------------------------------------------------------

# Cache of the already created database engines
# dictionary: database uri -> database engine

_engines = {}


def set_metadata(metadata, database_uri, database_debug, engine_settings):
    """Activate the metadatas (bind them to a database engine)

    In:
      - ``metadata`` -- the metadatas
      - ``database_uri`` -- connection string for the database engine
      - ``database_debug`` -- debug mode for the database engine
      - ``engine_settings`` -- dedicated parameters for the used database engine
    """
    if not metadata.bind:
        metadata.bind = _engines.setdefault(database_uri, sqlalchemy.engine_from_config(engine_settings, '', echo=database_debug, url=database_uri))
        setup_all()



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

import cStringIO
import cPickle

from nagare.continuation import Tasklet
from nagare.component import Component


def persistent_id(o, clean_callbacks, callbacks, session_data, tasklets):
    """An object with a ``_persistent_id`` attribute is stored into the session
    not into the state snapshot

    In:
      - ``o`` -- object to check
      - ``clean_callbacks`` -- do we have to forget the old callbacks?

    Out:
      - ``callbacks`` -- merge of the callbacks from all the components
      - ``session_data`` -- dict persistent_id -> object of the objects to store into the session
      - ``tasklets`` -- set of the serialized tasklets

    Return:
      - the persistent id or ``None``
    """
    r = None

    id_ = getattr(o, '_persistent_id', None)
    if id_ is not None:
        session_data[id_] = o
        r = str(id_)

    elif type(o) is Tasklet:
        tasklets.add(o)

    elif isinstance(o, Component):
        callbacks.update(o.serialize_callbacks(clean_callbacks))

    return r


def set_persistent_id(pickler, persistent_id):
    if 'inst_persistent_id' in dir(pickler):
        pickler.inst_persistent_id = persistent_id
    else:
        pickler.persistent_id = persistent_id


class DummyFile(object):
    """A write-only file that does nothing"""
    def write(self, data):
        pass


class Dummy(object):
    def __init__(self, pickler=None, unpickler=None):
        """Initialization

          - ``pickler`` -- pickler to use
          - ``unpickler`` -- unpickler to use
        """
        self.pickler = pickler or cPickle.Pickler
        self.unpickler = unpickler or cPickle.Unpickler

    def _dumps(self, pickler, data, clean_callbacks):
        """Serialize an objects graph

        In:
          - ``pickler`` -- pickler to use
          - ``data`` -- the objects graph
          - ``clean_callbacks`` -- do we have to forget the old callbacks?

        Out:
          - data to keep into the session
          - data to keep into the state
        """
        session_data = {}
        tasklets = set()
        callbacks = {}

        # Serialize the objects graph and extract all the callbacks
        set_persistent_id(pickler, lambda o: persistent_id(o, clean_callbacks, callbacks, session_data, tasklets))
        pickler.dump(data)

        return session_data, callbacks, tasklets

    def dumps(self, data, clean_callbacks):
        """Serialize an objects graph

        In:
          - ``data`` -- the objects graph
          - ``clean_callbacks`` -- do we have to forget the old callbacks?

        Out:
          - data kept into the session
          - data kept into the state
        """
        pickler = self.pickler(DummyFile(), protocol=-1)
        session_data, callbacks, tasklets = self._dumps(pickler, data, clean_callbacks)

        # This dummy serializer returns the data untouched
        return None, (data, callbacks)

    def loads(self, session_data, state_data):
        """Deserialize an objects graph

        In:
          - ``session_data`` -- data from the session
          - ``state_data`` -- data from the state

        Out:
          - the objects graph
          - the callbacks
        """
        return state_data


class Pickle(Dummy):
    def dumps(self, data, clean_callbacks):
        """Serialize an objects graph

        In:
          - ``data`` -- the objects graph
          - ``clean_callbacks`` -- do we have to forget the old callbacks?

        Out:
          - data kept into the session
          - data kept into the state
        """
        f = cStringIO.StringIO()
        pickler = self.pickler(f, protocol=-1)
        # Pickle the data
        session_data, callbacks, tasklets = self._dumps(pickler, data, clean_callbacks)

        # Pickle the callbacks
        set_persistent_id(pickler, lambda o: None)
        pickler.dump(callbacks)

        # Kill all the blocked tasklets, which are now serialized
        for t in tasklets:
            t.kill()

        # The pickled data are returned
        return session_data, f.getvalue()

    def loads(self, session_data, state_data):
        """Deserialize an objects graph

        In:
          - ``session_data`` -- data from the session
          - ``state_data`` -- data from the state

        Out:
          - the objects graph
          - the callbacks
        """
        p = self.unpickler(cStringIO.StringIO(state_data))
        if session_data:
            p.persistent_load = lambda i: session_data.get(int(i))

        return p.load(), p.load()



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Authentication manager for the basic HTTP authentication scheme"""

from base64 import b64decode

from webob import exc

from nagare.security import common


class Authentication(common.Authentication):
    """Authentication manager for the basic HTTP authentication scheme
    """
    def __init__(self, realm):
        """Initialization

        In:
          - ``realm`` -- authentication realm
        """
        self.realm = str(realm) if realm is not None else None

    def _get_ids(self, request, response):
        """Return the data associated with the connected user

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object

        Return:
          - A list with the id of the user and its password
        """
        ids = [None, None]  # The anonymous user by default

        authorization = request.headers.get('authorization')
        if authorization is not None:
            (scheme, data) = authorization.split(' ', 1)
            if scheme == 'Basic':
                encoding = request.accept_charset.best_match(['iso-8859-1', 'utf-8'])
                ids = [s.decode(encoding) for s in b64decode(data).split(':', 1)]

        return ids

    def get_ids(self, request, response):
        """Return the data associated with the connected user

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object

        Return:
          - A tuple with the id of the user and its password as a dictionary
        """
        (username, password) = self._get_ids(request, response)
        return username, {'password': password}

    def check_password(self, username, real_password, password):
        """Authentication

        In:
          - ``username`` -- user id
          - ``real_password`` -- real password of the user
          - ``password`` -- password received

        Return:
          - a boolean
        """
        return real_password == password

    def denies(self, detail):
        """Method called when a permission is denied

        In:
          - ``details`` -- a ``security.common.denial`` object
        """
        raise exc.HTTPUnauthorized(
            str(detail),
            [('WWW-Authenticate', 'Basic realm="%s"' % self.realm)]
        )



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

import unittest

from lxml import etree as ET
from webob import Request, Response

from nagare.namespaces import xhtml
from nagare import top


class TestTop(unittest.TestCase):
    def setUp(self):
        response = Response()
        response.xml_output = False
        self.h = xhtml.Renderer(request=Request({'PATH_INFO': ''}), response=response)

    def check_html(self, html):
        html2 = top.wrap('text/html', self.h, self.h.root)
        if not isinstance(html2, list):
            html2 = [html2]
        html2 = ''.join(map(ET.tostring, html2))

        self.assertEqual(html2, html)

    # Without <html>. Without <head>. Without <body>

    def test_1(self):
        self.h << 'hello'
        self.check_html('<html><head/><body>hello</body></html>')

    def test_2(self):
        h = self.h

        h << h.comment('c1') << 'hello' << h.comment('c2')
        self.check_html('<html><head/><body><!--c1-->hello<!--c2--></body></html>')

    def test_3(self):
        h = self.h

        h << h.p('hello')
        self.check_html('<html><head/><body><p>hello</p></body></html>')

    def test_4(self):
        h = self.h

        h << h.comment('c1') << h.p('hello') << h.comment('c2') << h.p('world') << h.comment('c3')
        self.check_html('<html><head/><body><!--c1--><p>hello</p><!--c2--><p>world</p><!--c3--></body></html>')

    # Without <html>. Without <head>. With <body>

    def test_5(self):
        h = self.h

        h << h.body('hello')
        self.check_html('<html><head/><body>hello</body></html>')

    def test_6(self):
        h = self.h

        h << h.comment('c1') << h.body('hello') << h.comment('c2')
        self.check_html('<html><head/><!--c1--><body>hello</body><!--c2--></html>')

    def test_7(self):
        h = self.h

        h << h.body(h.p('hello'))
        self.check_html('<html><head/><body><p>hello</p></body></html>')

    def test_8(self):
        h = self.h

        h << h.comment('c1') << h.body(h.p('hello')) << h.comment('c2')
        self.check_html('<html><head/><!--c1--><body><p>hello</p></body><!--c2--></html>')

    # Without <html>. With <head>. Without <body>

    def test_9(self):
        h = self.h

        h << h.head.head
        self.check_html('<html><head/><body/></html>')

    def test_10(self):
        h = self.h

        h << h.comment('c1') << h.head.head << h.comment('c2')
        self.check_html('<html><!--c1--><head/><body><!--c2--></body></html>')

    def test_11(self):
        h = self.h

        h << h.head.head << 'hello'
        self.check_html('<html><head/>hello<body/></html>')

    def test_12(self):
        h = self.h

        h << h.comment('c1') << h.head.head << h.p('hello') << h.comment('c2')
        self.check_html('<html><!--c1--><head/><body><p>hello</p><!--c2--></body></html>')

    def test_13(self):
        h = self.h

        h << h.comment('c1') << h.head.head << h.comment('c2') << h.p('hello') << h.comment('c3')
        self.check_html('<html><!--c1--><head/><body><!--c2--><p>hello</p><!--c3--></body></html>')

    # Without <html>. With <head>. With <body>

    def test_14(self):

        h = self.h

        h << h.head.head << h.body
        self.check_html('<html><head/><body/></html>')

    def test_15(self):
        h = self.h

        h << h.comment('c1') << h.head.head << h.comment('c2') << h.body << h.comment('c3')
        self.check_html('<html><!--c1--><head/><!--c2--><body/><!--c3--></html>')

    def test_16(self):
        h = self.h

        h << h.head.head << 'hello' << h.body << 'world'
        self.check_html('<html><head/>hello<body/>world</html>')

    def test_17(self):
        h = self.h

        h << h.comment('c1') << h.head.head << h.p('hello') << h.comment('c2') << h.body('world') << h.comment('c3')
        self.check_html('<html><!--c1--><head/><p>hello</p><!--c2--><body>world</body><!--c3--></html>')

    # With <html>. Without <head>. Without <body>

    def test_18(self):
        h = self.h
        h << h.html << 'hello'
        self.check_html('<html><head/><body/></html>hello')

    def test_19(self):
        h = self.h

        h << h.comment('c1') << h.html << 'hello' << h.comment('c2')
        self.check_html('<!--c1--><html><head/><body/></html>hello<!--c2-->')

    def test_20(self):
        h = self.h

        with h.html:
            h << h.p('hello')

        self.check_html('<html><head/><body><p>hello</p></body></html>')

    def test_21(self):
        h = self.h

        with h.html:
            h << h.comment('c1') << h.p('hello') << h.comment('c2') << h.p('world') << h.comment('c3')

        self.check_html('<html><head/><body><!--c1--><p>hello</p><!--c2--><p>world</p><!--c3--></body></html>')

    # With <html>. Without <head>. With <body>

    def test_22(self):
        h = self.h

        with h.html:
            h << h.body('hello')
        self.check_html('<html><head/><body>hello</body></html>')

    def test_23(self):
        h = self.h

        with h.html:
            h << h.comment('c1') << h.body('hello') << h.comment('c2')
        self.check_html('<html><head/><!--c1--><body>hello</body><!--c2--></html>')

    def test_24(self):
        h = self.h

        with h.html:
            h << h.body(h.p('hello'))
        self.check_html('<html><head/><body><p>hello</p></body></html>')

    def test_25(self):
        h = self.h

        with h.html:
            h << h.comment('c1') << h.body(h.p('hello')) << h.comment('c2')
        self.check_html('<html><head/><!--c1--><body><p>hello</p></body><!--c2--></html>')

    # With <html>. With <head>. Without <body>

    def test_26(self):
        h = self.h

        with h.html:
            h << h.head.head
        self.check_html('<html><head/><body/></html>')

    def test_27(self):
        h = self.h

        with h.html:
            h << h.comment('c1') << h.head.head << h.comment('c2')
        self.check_html('<html><!--c1--><head/><body><!--c2--></body></html>')

    def test_28(self):
        h = self.h

        with h.html:
            h << h.head.head << 'hello'
        self.check_html('<html><head/>hello<body/></html>')

    def test_29(self):
        h = self.h

        with h.html:
            h << h.comment('c1') << h.head.head << h.p('hello') << h.comment('c2')
        self.check_html('<html><!--c1--><head/><body><p>hello</p><!--c2--></body></html>')

    def test_30(self):
        h = self.h

        with h.html:
            h << h.comment('c1') << h.head.head << h.comment('c2') << h.p('hello') << h.comment('c3')
        self.check_html('<html><!--c1--><head/><body><!--c2--><p>hello</p><!--c3--></body></html>')

    # With <html>. With <head>. With <body>

    def test_31(self):
        h = self.h

        with h.html:
            h << h.head.head << h.body
        self.check_html('<html><head/><body/></html>')

    def test_32(self):
        h = self.h

        with h.html:
            h << h.comment('c1') << h.head.head << h.comment('c2') << h.body << h.comment('c3')
        self.check_html('<html><!--c1--><head/><!--c2--><body/><!--c3--></html>')

    def test_33(self):
        h = self.h

        with h.html:
            h << h.head.head << 'hello' << h.body << 'world'
        self.check_html('<html><head/>hello<body/>world</html>')

    def test_34(self):
        h = self.h

        with h.html:
            h << h.comment('c1') << h.head.head << h.p('hello') << h.comment('c2') << h.body('world') << h.comment('c3')
        self.check_html('<html><!--c1--><head/><p>hello</p><!--c2--><body>world</body><!--c3--></html>')

    # ---

    def test_35(self):
        h = self.h

        h.head << h.head.title('hello')
        h << h.p('world')
        self.check_html('<html><head><title>hello</title></head><body><p>world</p></body></html>')

    def test_36(self):
        h = self.h

        h.head << h.head.title('hello')
        h << h.head.head(h.head.title('world')) << h.p('foo')
        self.check_html('<html><head><title>world</title><title>hello</title></head><body><p>foo</p></body></html>')

    def test_37(self):
        h = self.h

        h.head << h.head.head(id='header_id')
        h << h.head.head(class_='header_class') << h.p('foo')
        self.check_html('<html><head class="header_class" id="header_id"/><body><p>foo</p></body></html>')

    # ---

    def test_38(self):
        h = self.h

        h << h.processing_instruction('hello') << h.p('foo') << h.processing_instruction('world')
        self.check_html('<html><head/><body><?hello ?><p>foo</p><?world ?></body></html>')

    def test_39(self):
        h = self.h

        h << h.processing_instruction('hello') << h.html(h.p('foo'))
        self.check_html('<?hello ?><html><head/><body><p>foo</p></body></html>')

    def test_canonical(self):
        h = xhtml.Renderer(request=Request({'PATH_INFO': ''}), response=self.h.response)
        self.assertEqual(top.wrap('text/html', h, h.root).write_xmlstring(), '<html><head/><body></body></html>')

        h = xhtml.Renderer(request=Request({'PATH_INFO': ''}), response=self.h.response)
        h.head << h.head.link(rel='canonical', href='/bar')
        self.assertEqual(top.wrap('text/html', h, h.root).write_xmlstring(), '<html><head><link href="/bar" rel="canonical"/></head><body></body></html>')

        h = xhtml.Renderer(request=Request({'PATH_INFO': '/foo'}), response=self.h.response)
        self.assertEqual(top.wrap('text/html', h, h.root).write_xmlstring(), '<html><head><link href="/foo" rel="canonical"/></head><body></body></html>')

        h = xhtml.Renderer(request=Request({'PATH_INFO': '/foo'}), response=self.h.response)
        h.head << h.head.link(rel='canonical', href='/bar')
        self.assertEqual(top.wrap('text/html', h, h.root).write_xmlstring(), '<html><head><link href="/bar" rel="canonical"/></head><body></body></html>')
